# Copyright 2016 Intel Corporation
# Copyright (c) 2012 NTT DOCOMO, INC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from ironic.common import boot_modes
from ironic.common.i18n import _


opts = [
    cfg.URIOpt('http_url',
               schemes=['http', 'https'],
               help=_("ironic-conductor node's HTTP server URL. "
                      "Example: http://192.1.2.3:8080")),
    cfg.StrOpt('http_root',
               default='/httpboot',
               help=_("ironic-conductor node's HTTP root path.")),
    cfg.StrOpt('image_server_auth_strategy',
               default='noauth',
               choices=[('noauth', _('No authentication')),
                        ('http_basic', _('HTTP Basic authentication'))],
               mutable=True,
               help=_("Used to select authentication strategy against the "
                       "image hosting HTTP(S) server. When set to http_basic "
                       "it enables HTTP(S) Basic Authentication. "
                       "Exception is thrown in case of missing credentials. "
                       "When this option has a valid value such as http_basic"
                       ", the same single set of credentials will be used "
                       "against all user-image sources! Currently only the "
                       "http_basic option has any functionality.")),
    cfg.StrOpt('image_server_user',
               mutable=True,
               help=_("Can be used by any authentication strategy that "
                      "requires username credential. Currently utilized by "
                      "the http_basic authentication strategy.")),
    cfg.StrOpt('image_server_password',
               mutable=True,
               secret=True,
               help=_("Can be used by any authentication strategy that "
                      "requires password credential. Currently utilized by "
                      "the http_basic authentication strategy.")),
    cfg.URIOpt('external_http_url',
               schemes=['http', 'https'],
               help=_("URL of the ironic-conductor node's HTTP server for "
                      "boot methods such as virtual media, "
                      "where images could be served outside of the "
                      "provisioning network. Does not apply when Swift is "
                      "used. Defaults to http_url.")),
    cfg.URIOpt('external_callback_url',
               schemes=['http', 'https'],
               help=_("Agent callback URL of the bare metal API for boot "
                      "methods such as virtual media, where images could be "
                      "served outside of the provisioning network. Defaults "
                      "to the configuration from [service_catalog].")),
    cfg.BoolOpt('enable_ata_secure_erase',
                default=True,
                mutable=True,
                help=_('Whether to support the use of ATA Secure Erase '
                       'during the cleaning process. Defaults to True.')),
    cfg.BoolOpt('enable_nvme_secure_erase',
                default=True,
                mutable=True,
                help=_('Whether to support the use of NVMe Secure Erase '
                       'during the cleaning process. Currently nvme-cli '
                       'format command is supported with user-data and '
                       'crypto modes, depending on device capabilities.'
                       'Defaults to True.')),
    cfg.IntOpt('erase_devices_priority',
               mutable=True,
               help=_('Priority to run in-band erase devices via the Ironic '
                      'Python Agent ramdisk. If unset, will use the priority '
                      'set in the ramdisk (defaults to 10 for the '
                      'GenericHardwareManager). If set to 0, will not run '
                      'during cleaning.')),
    cfg.IntOpt('erase_devices_metadata_priority',
               mutable=True,
               help=_('Priority to run in-band clean step that erases '
                      'metadata from devices, via the Ironic Python Agent '
                      'ramdisk. If unset, will use the priority set in the '
                      'ramdisk (defaults to 99 for the '
                      'GenericHardwareManager). If set to 0, will not run '
                      'during cleaning.')),
    cfg.IntOpt('delete_configuration_priority',
               mutable=True,
               help=_('Priority to run in-band clean step that erases '
                      'RAID configuration from devices, via the Ironic '
                      'Python Agent ramdisk. If unset, will use the '
                      'priority set in the ramdisk (defaults to 0 for the '
                      'GenericHardwareManager). If set to 0, will not run '
                      'during cleaning.')),
    cfg.IntOpt('create_configuration_priority',
               mutable=True,
               help=_('Priority to run in-band clean step that creates '
                      'RAID configuration from devices, via the Ironic '
                      'Python Agent ramdisk. If unset, will use the '
                      'priority set in the ramdisk (defaults to 0 for the '
                      'GenericHardwareManager). If set to 0, will not run '
                      'during cleaning.')),
    cfg.IntOpt('shred_random_overwrite_iterations',
               default=1,
               min=0,
               mutable=True,
               help=_('During shred, overwrite all block devices N times with '
                      'random data. This is only used if a device could not '
                      'be ATA Secure Erased. Defaults to 1.')),
    cfg.BoolOpt('shred_final_overwrite_with_zeros',
                default=True,
                mutable=True,
                help=_("Whether to write zeros to a node's block devices "
                       "after writing random data. This will write zeros to "
                       "the device even when "
                       "deploy.shred_random_overwrite_iterations is 0. This "
                       "option is only used if a device could not be ATA "
                       "Secure Erased. Defaults to True.")),
    cfg.BoolOpt('continue_if_disk_secure_erase_fails',
                default=False,
                mutable=True,
                help=_('Defines what to do if a secure erase operation '
                       '(NVMe or ATA) fails during cleaning in the Ironic '
                       'Python Agent. If False, the cleaning operation will '
                       'fail and the node will be put in ``clean failed`` '
                       'state. If True, shred will be invoked and cleaning '
                       'will continue.')),
    cfg.IntOpt('disk_erasure_concurrency',
               default=4,
               min=1,
               mutable=True,
               help=_('Defines the target pool size used by Ironic Python '
                      'Agent ramdisk to erase disk devices. The number of '
                      'threads created to erase disks will not exceed this '
                      'value or the number of disks to be erased.')),
    cfg.BoolOpt('power_off_after_deploy_failure',
                default=True,
                mutable=True,
                help=_('Whether to power off a node after deploy failure. '
                       'Defaults to True.')),
    cfg.StrOpt('default_boot_mode',
               choices=[(boot_modes.UEFI, _('UEFI boot mode')),
                        (boot_modes.LEGACY_BIOS, _('Legacy BIOS boot mode'))],
               default=boot_modes.UEFI,
               mutable=True,
               help=_('Default boot mode to use when no boot mode is '
                      'requested in node\'s driver_info, capabilities or '
                      'in the `instance_info` configuration. Currently the '
                      'default boot mode is "%(uefi)s", but it was '
                      '"%(bios)s" previously in Ironic. It is recommended '
                      'to set an explicit value for this option, and if the '
                      'setting or default differs from nodes, to ensure that '
                      'nodes are configured specifically for their desired '
                      'boot mode.') % {
                          'bios': boot_modes.LEGACY_BIOS,
                          'uefi': boot_modes.UEFI}),
    cfg.BoolOpt('configdrive_use_object_store',
                default=False,
                deprecated_group='conductor',
                deprecated_name='configdrive_use_swift',
                mutable=True,
                help=_('Whether to upload the config drive to object store. '
                       'Set this option to True to store config drive '
                       'in a swift endpoint.')),
    cfg.StrOpt('http_image_subdir',
               default='agent_images',
               help=_('The name of subdirectory under ironic-conductor '
                      'node\'s HTTP root path which is used to place instance '
                      'images for the direct deploy interface, when local '
                      'HTTP service is incorporated to provide instance image '
                      'instead of swift tempurls.')),
    cfg.BoolOpt('fast_track',
                default=False,
                mutable=True,
                help=_('Whether to allow deployment agents to perform lookup, '
                       'heartbeat operations during initial states of a '
                       'machine lifecycle and by-pass the normal setup '
                       'procedures for a ramdisk. This feature also enables '
                       'power operations which are part of deployment '
                       'processes to be bypassed if the ramdisk has performed '
                       'a heartbeat operation using the fast_track_timeout '
                       'setting.')),
    cfg.IntOpt('fast_track_timeout',
               default=300,
               min=0,
               max=300,
               mutable=True,
               help=_('Seconds for which the last heartbeat event is to be '
                      'considered valid for the purpose of a fast '
                      'track sequence. This setting should generally be '
                      'less than the number of seconds for "Power-On Self '
                      'Test" and typical ramdisk start-up. This value should '
                      'not exceed the [api]ramdisk_heartbeat_timeout '
                      'setting.')),
    cfg.BoolOpt('erase_skip_read_only',
                default=False,
                mutable=True,
                help=_('If the ironic-python-agent should skip read-only '
                       'devices when running the "erase_devices" clean step '
                       'where block devices are zeroed out. This requires '
                       'ironic-python-agent 6.0.0 or greater. By default '
                       'a read-only device will cause non-metadata based '
                       'cleaning operations to fail due to the possible '
                       'operational security risk of data being retained '
                       'between deployments of the bare metal node.')),
    cfg.StrOpt('ramdisk_image_download_source',
               choices=[('http', _('In case the ramdisk is already a bootable '
                                   'iso, using this option it will be '
                                   'directly provided by an external HTTP '
                                   'service using its full url.')),
                        ('local', _('This is the default behavior. '
                                    'The image is downloaded, prepared and '
                                    'cached locally, to be served from '
                                    'the conductor.')),
                        ('swift', _('Same as "http", but if the image '
                                    'is a Glance UUID, it is exposed via a '
                                    'Swift temporary URL.'))],
               default='local',
               mutable=True,
               help=_('Specifies whether a boot iso image should be served '
                      'from its own original location using the image source '
                      'url directly, or if ironic should cache the image on '
                      'the conductor and serve it from ironic\'s own http '
                      'server.')),
    cfg.StrOpt('iso_master_path',
               default='/var/lib/ironic/master_iso_images',
               help=_('On the ironic-conductor node, directory where master '
                      'ISO images are stored on disk. '
                      'Setting to the empty string disables image caching.')),
    cfg.IntOpt('iso_cache_size',
               default=20480,
               help=_('Maximum size (in MiB) of cache for master ISO images, '
                      'including those in use.')),
    # 10080 here is 1 week - 60*24*7. It is entirely arbitrary in the absence
    # of a facility to disable the ttl entirely.
    cfg.IntOpt('iso_cache_ttl',
               default=10080,
               help=_('Maximum TTL (in minutes) for old master ISO images in '
                      'cache.')),
]


def register_opts(conf):
    conf.register_opts(opts, group='deploy')



# Copyright 2014 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Ironic object test utilities."""

import functools
import inspect

from ironic.common import exception
from ironic.common.i18n import _
from ironic import objects
from ironic.objects import notification
from ironic.tests.unit.db import utils as db_utils


def check_keyword_arguments(func):
    @functools.wraps(func)
    def wrapper(**kw):
        obj_type = kw.pop('object_type')
        result = func(**kw)

        extra_args = set(kw) - set(result)
        if extra_args:
            raise exception.InvalidParameterValue(
                _("Unknown keyword arguments (%(extra)s) were passed "
                  "while creating a test %(object_type)s object.") %
                {"extra": ", ".join(extra_args),
                 "object_type": obj_type})

        return result

    return wrapper


def get_test_node(ctxt, **kw):
    """Return a Node object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    kw['object_type'] = 'node'
    get_db_node_checked = check_keyword_arguments(db_utils.get_test_node)
    db_node = get_db_node_checked(**kw)

    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del db_node['id']
    node = objects.Node(ctxt)
    for key in db_node:
        if key == 'traits':
            # convert list of strings to object
            raw_traits = db_node['traits']
            trait_list = []
            for raw_trait in raw_traits:
                trait = objects.Trait(ctxt, trait=raw_trait)
                trait_list.append(trait)
            node.traits = objects.TraitList(ctxt, objects=trait_list)
            node.traits.obj_reset_changes()
        else:
            setattr(node, key, db_node[key])
    return node


def create_test_node(ctxt, **kw):
    """Create and return a test node object.

    Create a node in the DB and return a Node object with appropriate
    attributes.
    """
    node = get_test_node(ctxt, **kw)
    node.create()
    return node


def get_test_port(ctxt, **kw):
    """Return a Port object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    kw['object_type'] = 'port'
    get_db_port_checked = check_keyword_arguments(
        db_utils.get_test_port)
    db_port = get_db_port_checked(**kw)

    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del db_port['id']
    port = objects.Port(ctxt)
    for key in db_port:
        setattr(port, key, db_port[key])
    return port


def create_test_port(ctxt, **kw):
    """Create and return a test port object.

    Create a port in the DB and return a Port object with appropriate
    attributes.
    """
    port = get_test_port(ctxt, **kw)
    port.create()
    return port


def get_test_chassis(ctxt, **kw):
    """Return a Chassis object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    kw['object_type'] = 'chassis'
    get_db_chassis_checked = check_keyword_arguments(
        db_utils.get_test_chassis)
    db_chassis = get_db_chassis_checked(**kw)

    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del db_chassis['id']
    chassis = objects.Chassis(ctxt)
    for key in db_chassis:
        setattr(chassis, key, db_chassis[key])
    return chassis


def create_test_chassis(ctxt, **kw):
    """Create and return a test chassis object.

    Create a chassis in the DB and return a Chassis object with appropriate
    attributes.
    """
    chassis = get_test_chassis(ctxt, **kw)
    chassis.create()
    return chassis


def get_test_portgroup(ctxt, **kw):
    """Return a Portgroup object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    kw['object_type'] = 'portgroup'
    get_db_port_group_checked = check_keyword_arguments(
        db_utils.get_test_portgroup)
    db_portgroup = get_db_port_group_checked(**kw)

    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del db_portgroup['id']
    portgroup = objects.Portgroup(ctxt)
    for key in db_portgroup:
        setattr(portgroup, key, db_portgroup[key])
    return portgroup


def create_test_portgroup(ctxt, **kw):
    """Create and return a test portgroup object.

    Create a portgroup in the DB and return a Portgroup object with appropriate
    attributes.
    """
    portgroup = get_test_portgroup(ctxt, **kw)
    portgroup.create()
    return portgroup


def get_test_volume_connector(ctxt, **kw):
    """Return a VolumeConnector object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    db_volume_connector = db_utils.get_test_volume_connector(**kw)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del db_volume_connector['id']
    volume_connector = objects.VolumeConnector(ctxt)
    for key in db_volume_connector:
        setattr(volume_connector, key, db_volume_connector[key])
    return volume_connector


def create_test_volume_connector(ctxt, **kw):
    """Create and return a test volume connector object.

    Create a volume connector in the DB and return a VolumeConnector object
    with appropriate attributes.
    """
    volume_connector = get_test_volume_connector(ctxt, **kw)
    volume_connector.create()
    return volume_connector


def get_test_volume_target(ctxt, **kw):
    """Return a VolumeTarget object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    db_volume_target = db_utils.get_test_volume_target(**kw)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del db_volume_target['id']
    volume_target = objects.VolumeTarget(ctxt)
    for key in db_volume_target:
        setattr(volume_target, key, db_volume_target[key])
    return volume_target


def create_test_volume_target(ctxt, **kw):
    """Create and return a test volume target object.

    Create a volume target in the DB and return a VolumeTarget object with
    appropriate attributes.
    """
    volume_target = get_test_volume_target(ctxt, **kw)
    volume_target.create()
    return volume_target


def get_test_bios_setting(ctxt, **kw):
    """Return a BiosSettingList object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    kw['object_type'] = 'bios'
    db_bios_setting = db_utils.get_test_bios_setting(**kw)
    bios_setting = objects.BIOSSetting(ctxt)
    for key in db_bios_setting:
        setattr(bios_setting, key, db_bios_setting[key])
    return bios_setting


def create_test_bios_setting(ctxt, **kw):
    """Create and return a test bios setting list object.

    Create a BIOS setting list in the DB and return a BIOSSettingList
    object with appropriate attributes.
    """
    bios_setting = get_test_bios_setting(ctxt, **kw)
    bios_setting.create()
    return bios_setting


def create_test_conductor(ctxt, **kw):
    """Register and return a test conductor object."""
    args = db_utils.get_test_conductor(**kw)
    conductor = objects.Conductor.register(ctxt, args['hostname'],
                                           args['drivers'],
                                           args['conductor_group'],
                                           update_existing=True)
    return conductor


def get_test_allocation(ctxt, **kw):
    """Return an Allocation object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    kw['object_type'] = 'allocation'
    get_db_allocation_checked = check_keyword_arguments(
        db_utils.get_test_allocation)
    db_allocation = get_db_allocation_checked(**kw)

    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del db_allocation['id']
    allocation = objects.Allocation(ctxt)
    for key in db_allocation:
        setattr(allocation, key, db_allocation[key])
    return allocation


def create_test_allocation(ctxt, **kw):
    """Create and return a test allocation object.

    Create an allocation in the DB and return an Allocation object with
    appropriate attributes.
    """
    allocation = get_test_allocation(ctxt, **kw)
    allocation.create()
    return allocation


def get_test_deploy_template(ctxt, **kw):
    """Return a DeployTemplate object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    db_template = db_utils.get_test_deploy_template(**kw)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del db_template['id']
    if 'steps' not in kw:
        for step in db_template['steps']:
            del step['id']
            del step['deploy_template_id']
    else:
        for kw_step, template_step in zip(kw['steps'], db_template['steps']):
            if 'id' not in kw_step and 'id' in template_step:
                del template_step['id']
    template = objects.DeployTemplate(ctxt)
    for key in db_template:
        setattr(template, key, db_template[key])
    return template


def create_test_deploy_template(ctxt, **kw):
    """Create and return a test deploy template object.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    template = get_test_deploy_template(ctxt, **kw)
    template.create()
    return template


def get_payloads_with_schemas(from_module):
    """Get the Payload classes with SCHEMAs defined.

    :param from_module: module from which to get the classes.
    :returns: list of Payload classes that have SCHEMAs defined.

    """
    payloads = []
    for name, payload in inspect.getmembers(from_module, inspect.isclass):
        # Assume that Payload class names end in 'Payload'.
        if name.endswith("Payload"):
            base_classes = inspect.getmro(payload)
            if notification.NotificationPayloadBase not in base_classes:
                # The class may have the desired name but it isn't a REAL
                # Payload class; skip it.
                continue

            # First class is this payload class, parent class is the 2nd
            # one in the tuple
            parent = base_classes[1]
            if (not hasattr(parent, 'SCHEMA')
                or parent.SCHEMA != payload.SCHEMA):
                payloads.append(payload)

    return payloads


def get_test_runbook(ctxt, **kw):
    """Return a Runbook object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    db_runbook = db_utils.get_test_runbook(**kw)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del db_runbook['id']
    if 'steps' not in kw:
        for step in db_runbook['steps']:
            del step['id']
            del step['runbook_id']
    else:
        for kw_step, runbook_step in zip(kw['steps'], db_runbook['steps']):
            if 'id' not in kw_step and 'id' in runbook_step:
                del runbook_step['id']
    runbook = objects.Runbook(ctxt)
    for key in db_runbook:
        setattr(runbook, key, db_runbook[key])
    return runbook


def create_test_runbook(ctxt, **kw):
    """Create and return a test runbook object.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    runbook = get_test_runbook(ctxt, **kw)
    runbook.create()
    return runbook


class SchemasTestMixIn(object):
    def _check_payload_schemas(self, from_module, fields):
        """Assert that the Payload SCHEMAs have the expected properties.

           A payload's SCHEMA should:

           1. Have each of its keys in the payload's fields
           2. Have each member of the schema match with a corresponding field
           in the object
        """
        resource = from_module.__name__.split('.')[-1]
        payloads = get_payloads_with_schemas(from_module)
        for payload in payloads:
            for schema_key in payload.SCHEMA:
                self.assertIn(schema_key, payload.fields,
                              "for %s, schema key %s is not in fields"
                              % (payload, schema_key))
                key = payload.SCHEMA[schema_key][1]
                self.assertIn(key, fields,
                              "for %s, schema key %s has invalid %s "
                              "field %s" % (payload, schema_key, resource,
                                            key))


def create_test_inventory(ctxt, node, **kw):
    """Create and return a test node inventory object."""
    inv = objects.NodeInventory(ctxt)
    if not isinstance(node, str):
        node = node.id
    kw['node_id'] = node
    for key, value in kw.items():
        setattr(inv, key, value)
    inv.create()
    return inv


def get_test_firmware_component(ctxt, **kw):
    """Return a FirmwareComponent object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    kw['object_type'] = 'firmware'
    get_db_fw_checked = check_keyword_arguments(
        db_utils.get_test_firmware_component)
    db_fw_cmp = get_db_fw_checked(**kw)

    if 'id' not in kw:
        del db_fw_cmp['id']

    fw_cmp = objects.FirmwareComponent(ctxt)
    for key in db_fw_cmp:
        setattr(fw_cmp, key, db_fw_cmp[key])

    return fw_cmp


def create_test_firmware_component(ctxt, **kw):
    """Create and return a test Firmware Component object.

    Create a Firmware Component in the DB and return a FirmwareComponent
    object with appropriate attributes.
    """
    fw_cmp = get_test_firmware_component(ctxt, **kw)
    fw_cmp.create()
    return fw_cmp


def get_test_inspection_rule(ctxt, **kw):
    """Return an InpsectionRule object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    db_inspection_rule = db_utils.get_test_inspection_rule(**kw)
    inspection_rule = objects.InspectionRule(ctxt)
    for key in db_inspection_rule:
        setattr(inspection_rule, key, db_inspection_rule[key])
    return inspection_rule


def create_test_inspection_rule(ctxt, **kw):
    """Create and return a test inspection rule object.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    inspection_rule = get_test_inspection_rule(ctxt, **kw)
    inspection_rule.create()
    return inspection_rule



# Copyright 2011 Justin Santa Barbara
# Copyright 2012 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import datetime
import errno
import os
import os.path
import shutil
import tempfile
import time
from unittest import mock

import jinja2
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_utils import netutils
import psutil

from ironic.common import exception
from ironic.common import utils
from ironic.tests import base

CONF = cfg.CONF


class BareMetalUtilsTestCase(base.TestCase):

    def test_create_link(self):
        with mock.patch.object(os, "symlink", autospec=True) as symlink_mock:
            symlink_mock.return_value = None
            utils.create_link_without_raise("/fake/source", "/fake/link")
            symlink_mock.assert_called_once_with("/fake/source", "/fake/link")

    def test_create_link_EEXIST(self):
        with mock.patch.object(os, "symlink", autospec=True) as symlink_mock:
            symlink_mock.side_effect = OSError(errno.EEXIST)
            utils.create_link_without_raise("/fake/source", "/fake/link")
            symlink_mock.assert_called_once_with("/fake/source", "/fake/link")


class GenericUtilsTestCase(base.TestCase):
    @mock.patch.object(utils, 'hashlib', autospec=True)
    def test__get_hash_object(self, hashlib_mock):
        algorithms_available = ('md5', 'sha1', 'sha224',
                                'sha256', 'sha384', 'sha512')
        hashlib_mock.algorithms_guaranteed = algorithms_available
        hashlib_mock.algorithms = algorithms_available
        # | WHEN |
        utils._get_hash_object('md5')
        utils._get_hash_object('sha1')
        utils._get_hash_object('sha224')
        utils._get_hash_object('sha256')
        utils._get_hash_object('sha384')
        utils._get_hash_object('sha512')
        # | THEN |
        calls = [mock.call.md5(), mock.call.sha1(), mock.call.sha224(),
                 mock.call.sha256(), mock.call.sha384(), mock.call.sha512()]
        hashlib_mock.assert_has_calls(calls)

    def test__get_hash_object_throws_for_invalid_or_unsupported_hash_name(
            self):
        # | WHEN | & | THEN |
        self.assertRaises(exception.InvalidParameterValue,
                          utils._get_hash_object,
                          'hickory-dickory-dock')

    def test_file_has_content_equal(self):
        data = b'Mary had a little lamb, its fleece as white as snow'
        ref = data
        with mock.patch('oslo_utils.fileutils.open',
                        mock.mock_open(read_data=data)) as mopen:
            self.assertTrue(utils.file_has_content('foo', ref))
            mopen.assert_called_once_with('foo', 'rb')

    def test_file_has_content_equal_not_binary(self):
        data = ('Mary had a little lamb, its fleece as white as '
                'sno\u0449').encode('utf-8')
        ref = data
        with mock.patch('oslo_utils.fileutils.open',
                        mock.mock_open(read_data=data)) as mopen:
            self.assertTrue(utils.file_has_content('foo', ref))
            mopen.assert_called_once_with('foo', 'rb')

    def test_file_has_content_differ(self):
        data = b'Mary had a little lamb, its fleece as white as snow'
        ref = data + b'!'
        with mock.patch('oslo_utils.fileutils.open',
                        mock.mock_open(read_data=data)) as mopen:
            self.assertFalse(utils.file_has_content('foo', ref))
            mopen.assert_called_once_with('foo', 'rb')

    def test_is_valid_datapath_id(self):
        self.assertTrue(utils.is_valid_datapath_id("525400cf2d319fdf"))
        self.assertTrue(utils.is_valid_datapath_id("525400CF2D319FDF"))
        self.assertFalse(utils.is_valid_datapath_id("52"))
        self.assertFalse(utils.is_valid_datapath_id("52:54:00:cf:2d:31"))
        self.assertFalse(utils.is_valid_datapath_id("notadatapathid00"))
        self.assertFalse(utils.is_valid_datapath_id("5525400CF2D319FDF"))

    def test_is_hostname_safe(self):
        self.assertTrue(utils.is_hostname_safe('spam'))
        self.assertFalse(utils.is_hostname_safe('spAm'))
        self.assertFalse(utils.is_hostname_safe('SPAM'))
        self.assertFalse(utils.is_hostname_safe('-spam'))
        self.assertFalse(utils.is_hostname_safe('spam-'))
        self.assertTrue(utils.is_hostname_safe('spam-eggs'))
        self.assertFalse(utils.is_hostname_safe('spam_eggs'))
        self.assertFalse(utils.is_hostname_safe('spam eggs'))
        self.assertTrue(utils.is_hostname_safe('spam.eggs'))
        self.assertTrue(utils.is_hostname_safe('9spam'))
        self.assertTrue(utils.is_hostname_safe('spam7'))
        self.assertTrue(utils.is_hostname_safe('br34kf4st'))
        self.assertFalse(utils.is_hostname_safe('$pam'))
        self.assertFalse(utils.is_hostname_safe('egg$'))
        self.assertFalse(utils.is_hostname_safe('spam#eggs'))
        self.assertFalse(utils.is_hostname_safe(' eggs'))
        self.assertFalse(utils.is_hostname_safe('spam '))
        self.assertTrue(utils.is_hostname_safe('s'))
        self.assertTrue(utils.is_hostname_safe('s' * 63))
        self.assertFalse(utils.is_hostname_safe('s' * 64))
        self.assertFalse(utils.is_hostname_safe(''))
        self.assertFalse(utils.is_hostname_safe(None))
        # Need to ensure a binary response for success or fail
        self.assertIsNotNone(utils.is_hostname_safe('spam'))
        self.assertIsNotNone(utils.is_hostname_safe('-spam'))
        self.assertTrue(utils.is_hostname_safe('www.rackspace.com'))
        self.assertTrue(utils.is_hostname_safe('www.rackspace.com.'))
        self.assertTrue(utils.is_hostname_safe('http._sctp.www.example.com'))
        self.assertTrue(utils.is_hostname_safe('mail.pets_r_us.net'))
        self.assertTrue(utils.is_hostname_safe('mail-server-15.my_host.org'))
        self.assertFalse(utils.is_hostname_safe('www.nothere.com_'))
        self.assertFalse(utils.is_hostname_safe('www.nothere_.com'))
        self.assertFalse(utils.is_hostname_safe('www..nothere.com'))
        long_str = 'a' * 63 + '.' + 'b' * 63 + '.' + 'c' * 63 + '.' + 'd' * 63
        self.assertTrue(utils.is_hostname_safe(long_str))
        self.assertFalse(utils.is_hostname_safe(long_str + '.'))
        self.assertFalse(utils.is_hostname_safe('a' * 255))

    def test_is_valid_logical_name(self):
        valid = (
            'spam', 'spAm', 'SPAM', 'spam-eggs', 'spam.eggs', 'spam_eggs',
            'spam~eggs', '9spam', 'spam7', '~spam', '.spam', '.~-_', '~',
            'br34kf4st', 's', 's' * 63, 's' * 255)
        invalid = (
            ' ', 'spam eggs', '$pam', 'egg$', 'spam#eggs',
            ' eggs', 'spam ', '', None, 'spam%20')

        for hostname in valid:
            result = utils.is_valid_logical_name(hostname)
            # Need to ensure a binary response for success. assertTrue
            # is too generous, and would pass this test if, for
            # instance, a regex Match object were returned.
            self.assertIs(result, True,
                          "%s is unexpectedly invalid" % hostname)

        for hostname in invalid:
            result = utils.is_valid_logical_name(hostname)
            # Need to ensure a binary response for
            # success. assertFalse is too generous and would pass this
            # test if None were returned.
            self.assertIs(result, False,
                          "%s is unexpectedly valid" % hostname)

    def test_validate_and_normalize_mac(self):
        mac = 'AA:BB:CC:DD:EE:FF'
        with mock.patch.object(netutils, 'is_valid_mac',
                               autospec=True) as m_mock:
            m_mock.return_value = True
            self.assertEqual(mac.lower(),
                             utils.validate_and_normalize_mac(mac))

    def test_validate_and_normalize_datapath_id(self):
        datapath_id = 'AA:BB:CC:DD:EE:FF'
        with mock.patch.object(utils, 'is_valid_datapath_id',
                               autospec=True) as m_mock:
            m_mock.return_value = True
            self.assertEqual(datapath_id.lower(),
                             utils.validate_and_normalize_datapath_id(
                                 datapath_id))

    def test_validate_and_normalize_mac_invalid_format(self):
        with mock.patch.object(netutils, 'is_valid_mac',
                               autospec=True) as m_mock:
            m_mock.return_value = False
            self.assertRaises(exception.InvalidMAC,
                              utils.validate_and_normalize_mac, 'invalid-mac')

    def test_safe_rstrip(self):
        value = '/test/'
        rstripped_value = '/test'
        not_rstripped = '/'

        self.assertEqual(rstripped_value, utils.safe_rstrip(value, '/'))
        self.assertEqual(not_rstripped, utils.safe_rstrip(not_rstripped, '/'))

    def test_safe_rstrip_not_raises_exceptions(self):
        # Supplying an integer should normally raise an exception because it
        # does not save the rstrip() method.
        value = 10

        # In the case of raising an exception safe_rstrip() should return the
        # original value.
        self.assertEqual(value, utils.safe_rstrip(value))

    @mock.patch.object(os.path, 'getmtime', return_value=1439465889.4964755,
                       autospec=True)
    def test_unix_file_modification_datetime(self, mtime_mock):
        expected = datetime.datetime(2015, 8, 13, 11, 38, 9, 496475)
        self.assertEqual(expected,
                         utils.unix_file_modification_datetime('foo'))
        mtime_mock.assert_called_once_with('foo')

    def test_is_valid_no_proxy(self):

        # Valid values for 'no_proxy'
        valid_no_proxy = [
            ('a' * 63 + '.' + '0' * 63 + '.c.' + 'd' * 61 + '.' + 'e' * 61),
            ('A' * 63 + '.' + '0' * 63 + '.C.' + 'D' * 61 + '.' + 'E' * 61),
            ('.' + 'a' * 62 + '.' + '0' * 62 + '.c.' + 'd' * 61 + '.'
             + 'e' * 61),
            ',,example.com:3128,',
            '192.168.1.1',  # IP should be valid
        ]
        # Test each one individually, so if failure easier to determine which
        # one failed.
        for no_proxy in valid_no_proxy:
            self.assertTrue(
                utils.is_valid_no_proxy(no_proxy),
                msg="'no_proxy' value should be valid: {}".format(no_proxy))
        # Test valid when joined together
        self.assertTrue(utils.is_valid_no_proxy(','.join(valid_no_proxy)))
        # Test valid when joined together with whitespace
        self.assertTrue(utils.is_valid_no_proxy(' , '.join(valid_no_proxy)))
        # empty string should also be valid
        self.assertTrue(utils.is_valid_no_proxy(''))

        # Invalid values for 'no_proxy'
        invalid_no_proxy = [
            ('A' * 64 + '.' + '0' * 63 + '.C.' + 'D' * 61 + '.'
             + 'E' * 61),  # too long (> 253)
            ('a' * 100),
            'a..com',
            ('.' + 'a' * 63 + '.' + '0' * 62 + '.c.' + 'd' * 61 + '.'
             + 'e' * 61),  # too long (> 251 after deleting .)
            ('*.' + 'a' * 60 + '.' + '0' * 60 + '.c.' + 'd' * 61 + '.'
             + 'e' * 61),  # starts with *.
            'c.-a.com',
            'c.a-.com',
        ]

        for no_proxy in invalid_no_proxy:
            self.assertFalse(
                utils.is_valid_no_proxy(no_proxy),
                msg="'no_proxy' value should be invalid: {}".format(no_proxy))

    def test_is_fips_enabled(self):
        with mock.patch('builtins.open', mock.mock_open(read_data='1\n')) as m:
            self.assertTrue(utils.is_fips_enabled())
            m.assert_called_once_with('/proc/sys/crypto/fips_enabled', 'r')

        with mock.patch('builtins.open', mock.mock_open(read_data='0\n')) as m:
            self.assertFalse(utils.is_fips_enabled())
            m.assert_called_once_with('/proc/sys/crypto/fips_enabled', 'r')

        mock_open = mock.mock_open()
        mock_open.side_effect = FileNotFoundError
        with mock.patch('builtins.open', mock_open) as m:
            self.assertFalse(utils.is_fips_enabled())
            m.assert_called_once_with('/proc/sys/crypto/fips_enabled', 'r')

    def test_wrap_ipv6(self):
        self.assertEqual('[2001:DB8::1]', utils.wrap_ipv6('2001:DB8::1'))
        self.assertEqual('example.com', utils.wrap_ipv6('example.com'))
        self.assertEqual('192.168.24.1', utils.wrap_ipv6('192.168.24.1'))
        self.assertEqual('[2001:DB8::1]', utils.wrap_ipv6('[2001:DB8::1]'))


class TempFilesTestCase(base.TestCase):

    def test_tempdir(self):

        dirname = None
        with utils.tempdir() as tempdir:
            self.assertTrue(os.path.isdir(tempdir))
            dirname = tempdir
        self.assertFalse(os.path.exists(dirname))

    @mock.patch.object(shutil, 'rmtree', autospec=True)
    @mock.patch.object(tempfile, 'mkdtemp', autospec=True)
    def test_tempdir_mocked(self, mkdtemp_mock, rmtree_mock):

        self.config(tempdir='abc')
        mkdtemp_mock.return_value = 'temp-dir'
        kwargs = {'dir': 'b'}

        with utils.tempdir(**kwargs) as tempdir:
            self.assertEqual('temp-dir', tempdir)
            tempdir_created = tempdir

        mkdtemp_mock.assert_called_once_with(**kwargs)
        rmtree_mock.assert_called_once_with(tempdir_created)

    @mock.patch.object(utils, 'LOG', autospec=True)
    @mock.patch.object(shutil, 'rmtree', autospec=True)
    @mock.patch.object(tempfile, 'mkdtemp', autospec=True)
    def test_tempdir_mocked_error_on_rmtree(self, mkdtemp_mock, rmtree_mock,
                                            log_mock):

        self.config(tempdir='abc')
        mkdtemp_mock.return_value = 'temp-dir'
        rmtree_mock.side_effect = OSError

        with utils.tempdir() as tempdir:
            self.assertEqual('temp-dir', tempdir)
            tempdir_created = tempdir

        rmtree_mock.assert_called_once_with(tempdir_created)
        self.assertTrue(log_mock.error.called)

    @mock.patch.object(os.path, 'exists', autospec=True)
    @mock.patch.object(utils, '_check_dir_writable', autospec=True)
    @mock.patch.object(utils, '_check_dir_free_space', autospec=True)
    def test_check_dir_with_pass_in(self, mock_free_space, mock_dir_writable,
                                    mock_exists):
        mock_exists.return_value = True
        # test passing in a directory and size
        utils.check_dir(directory_to_check='/fake/path', required_space=5)
        mock_exists.assert_called_once_with('/fake/path')
        mock_dir_writable.assert_called_once_with('/fake/path')
        mock_free_space.assert_called_once_with('/fake/path', 5)

    @mock.patch.object(utils, '_check_dir_writable', autospec=True)
    @mock.patch.object(utils, '_check_dir_free_space', autospec=True)
    def test_check_dir_no_dir(self, mock_free_space, mock_dir_writable):
        self.config(tempdir='/fake/path')
        # NOTE(dtantsur): self.config uses os.path.exists, so we cannot mock
        # on the method level.
        with mock.patch.object(os.path, 'exists',
                               autospec=True) as mock_exists:
            mock_exists.return_value = False
            self.assertRaises(exception.PathNotFound, utils.check_dir)
            mock_exists.assert_called_once_with(CONF.tempdir)
        self.assertFalse(mock_free_space.called)
        self.assertFalse(mock_dir_writable.called)

    @mock.patch.object(utils, '_check_dir_writable', autospec=True)
    @mock.patch.object(utils, '_check_dir_free_space', autospec=True)
    def test_check_dir_ok(self, mock_free_space, mock_dir_writable):
        self.config(tempdir='/fake/path')
        # NOTE(dtantsur): self.config uses os.path.exists, so we cannot mock
        # on the method level.
        with mock.patch.object(os.path, 'exists',
                               autospec=True) as mock_exists:
            mock_exists.return_value = True
            utils.check_dir()
            mock_exists.assert_called_once_with(CONF.tempdir)
        mock_dir_writable.assert_called_once_with(CONF.tempdir)
        mock_free_space.assert_called_once_with(CONF.tempdir, 1)

    @mock.patch.object(os, 'access', autospec=True)
    def test__check_dir_writable_ok(self, mock_access):
        mock_access.return_value = True
        self.assertIsNone(utils._check_dir_writable("/fake/path"))
        mock_access.assert_called_once_with("/fake/path", os.W_OK)

    @mock.patch.object(os, 'access', autospec=True)
    def test__check_dir_writable_not_writable(self, mock_access):
        mock_access.return_value = False

        self.assertRaises(exception.DirectoryNotWritable,
                          utils._check_dir_writable, "/fake/path")
        mock_access.assert_called_once_with("/fake/path", os.W_OK)

    @mock.patch.object(os, 'statvfs', autospec=True)
    def test__check_dir_free_space_ok(self, mock_stat):
        statvfs_mock_return = mock.MagicMock()
        statvfs_mock_return.f_bsize = 5
        statvfs_mock_return.f_frsize = 0
        statvfs_mock_return.f_blocks = 0
        statvfs_mock_return.f_bfree = 0
        statvfs_mock_return.f_bavail = 1024 * 1024
        statvfs_mock_return.f_files = 0
        statvfs_mock_return.f_ffree = 0
        statvfs_mock_return.f_favail = 0
        statvfs_mock_return.f_flag = 0
        statvfs_mock_return.f_namemax = 0
        mock_stat.return_value = statvfs_mock_return
        utils._check_dir_free_space("/fake/path")
        mock_stat.assert_called_once_with("/fake/path")

    @mock.patch.object(os, 'statvfs', autospec=True)
    def test_check_dir_free_space_raises(self, mock_stat):
        statvfs_mock_return = mock.MagicMock()
        statvfs_mock_return.f_bsize = 1
        statvfs_mock_return.f_frsize = 0
        statvfs_mock_return.f_blocks = 0
        statvfs_mock_return.f_bfree = 0
        statvfs_mock_return.f_bavail = 1024
        statvfs_mock_return.f_files = 0
        statvfs_mock_return.f_ffree = 0
        statvfs_mock_return.f_favail = 0
        statvfs_mock_return.f_flag = 0
        statvfs_mock_return.f_namemax = 0
        mock_stat.return_value = statvfs_mock_return

        self.assertRaises(exception.InsufficientDiskSpace,
                          utils._check_dir_free_space, "/fake/path")
        mock_stat.assert_called_once_with("/fake/path")

    @mock.patch.object(time, 'sleep', autospec=True)
    @mock.patch.object(psutil, 'virtual_memory', autospec=True)
    def test_is_memory_insufficient(self, mock_vm_check, mock_sleep):

        class vm_check(object):
            available = 1000000000

        mock_vm_check.return_value = vm_check
        self.assertTrue(utils.is_memory_insufficient())
        self.assertEqual(14, mock_vm_check.call_count)

    @mock.patch.object(time, 'sleep', autospec=True)
    @mock.patch.object(psutil, 'virtual_memory', autospec=True)
    def test_is_memory_insufficient_good(self, mock_vm_check,
                                         mock_sleep):

        class vm_check(object):
            available = 3276700000

        mock_vm_check.return_value = vm_check
        self.assertFalse(utils.is_memory_insufficient())
        self.assertEqual(1, mock_vm_check.call_count)

    @mock.patch.object(time, 'sleep', autospec=True)
    @mock.patch.object(psutil, 'virtual_memory', autospec=True)
    def test_is_memory_insufficient_recovers(self, mock_vm_check,
                                             mock_sleep):

        class vm_check_bad(object):
            available = 1023000000

        class vm_check_good(object):
            available = 3276700000

        self.config(minimum_memory_warning_only=False)
        mock_vm_check.side_effect = iter([vm_check_bad,
                                          vm_check_bad,
                                          vm_check_good])
        self.assertFalse(utils.is_memory_insufficient())
        self.assertEqual(3, mock_vm_check.call_count)

    @mock.patch.object(time, 'sleep', autospec=True)
    @mock.patch.object(psutil, 'virtual_memory', autospec=True)
    def test_is_memory_insufficient_warning_only(self, mock_vm_check,
                                                 mock_sleep):
        self.config(minimum_memory_warning_only=True)

        class vm_check_bad(object):
            available = 1023000000

        mock_vm_check.side_effect = vm_check_bad
        self.assertFalse(utils.is_memory_insufficient())
        self.assertEqual(2, mock_vm_check.call_count)


class GetUpdatedCapabilitiesTestCase(base.TestCase):

    def test_get_updated_capabilities(self):
        capabilities = {'ilo_firmware_version': 'xyz'}
        cap_string = 'ilo_firmware_version:xyz'
        cap_returned = utils.get_updated_capabilities(None, capabilities)
        self.assertEqual(cap_string, cap_returned)
        self.assertIsInstance(cap_returned, str)

    def test_get_updated_capabilities_multiple_keys(self):
        capabilities = {'ilo_firmware_version': 'xyz',
                        'foo': 'bar', 'somekey': 'value'}
        cap_string = 'ilo_firmware_version:xyz,foo:bar,somekey:value'
        cap_returned = utils.get_updated_capabilities(None, capabilities)
        set1 = set(cap_string.split(','))
        set2 = set(cap_returned.split(','))
        self.assertEqual(set1, set2)
        self.assertIsInstance(cap_returned, str)

    def test_get_updated_capabilities_invalid_capabilities(self):
        capabilities = 'ilo_firmware_version'
        self.assertRaises(ValueError,
                          utils.get_updated_capabilities,
                          capabilities, {})

    def test_get_updated_capabilities_capabilities_not_dict(self):
        capabilities = ['ilo_firmware_version:xyz', 'foo:bar']
        self.assertRaises(ValueError,
                          utils.get_updated_capabilities,
                          None, capabilities)

    def test_get_updated_capabilities_add_to_existing_capabilities(self):
        new_capabilities = {'BootMode': 'uefi'}
        expected_capabilities = 'BootMode:uefi,foo:bar'
        cap_returned = utils.get_updated_capabilities('foo:bar',
                                                      new_capabilities)
        set1 = set(expected_capabilities.split(','))
        set2 = set(cap_returned.split(','))
        self.assertEqual(set1, set2)
        self.assertIsInstance(cap_returned, str)

    def test_get_updated_capabilities_replace_to_existing_capabilities(self):
        new_capabilities = {'BootMode': 'bios'}
        expected_capabilities = 'BootMode:bios'
        cap_returned = utils.get_updated_capabilities('BootMode:uefi',
                                                      new_capabilities)
        set1 = set(expected_capabilities.split(','))
        set2 = set(cap_returned.split(','))
        self.assertEqual(set1, set2)
        self.assertIsInstance(cap_returned, str)

    def test_validate_network_port(self):
        port = utils.validate_network_port('0', 'message')
        self.assertEqual(0, port)
        port = utils.validate_network_port('65535')
        self.assertEqual(65535, port)

    def test_validate_network_port_fail(self):
        self.assertRaisesRegex(exception.InvalidParameterValue,
                               'Port "65536" is not a valid port.',
                               utils.validate_network_port,
                               '65536')
        self.assertRaisesRegex(exception.InvalidParameterValue,
                               'fake_port "-1" is not a valid port.',
                               utils.validate_network_port,
                               '-1',
                               'fake_port')
        self.assertRaisesRegex(exception.InvalidParameterValue,
                               'Port "invalid" is not a valid port.',
                               utils.validate_network_port,
                               'invalid')


class JinjaTemplatingTestCase(base.TestCase):

    def setUp(self):
        super(JinjaTemplatingTestCase, self).setUp()
        self.template = '{{ foo }} {{ bar }}'
        self.params = {'foo': 'spam', 'bar': 'ham'}
        self.expected = 'spam ham'

    def test_render_string(self):
        self.assertEqual(self.expected,
                         utils.render_template(self.template,
                                               self.params,
                                               is_file=False))

    def test_render_with_quotes(self):
        """test jinja2 autoescaping for everything is disabled """
        self.expected = '"spam" ham'
        self.params = {'foo': '"spam"', 'bar': 'ham'}
        self.assertEqual(self.expected,
                         utils.render_template(self.template,
                                               self.params,
                                               is_file=False))

    @mock.patch('ironic.common.utils.jinja2.FileSystemLoader', autospec=True)
    def test_render_file(self, jinja_fsl_mock):
        path = '/path/to/template.j2'
        jinja_fsl_mock.return_value = jinja2.DictLoader(
            {'template.j2': self.template})
        self.assertEqual(self.expected,
                         utils.render_template(path,
                                               self.params))
        jinja_fsl_mock.assert_called_once_with('/path/to')


class ValidateConductorGroupTestCase(base.TestCase):
    def test_validate_conductor_group_success(self):
        self.assertIsNone(utils.validate_conductor_group('foo'))
        self.assertIsNone(utils.validate_conductor_group('group1'))
        self.assertIsNone(utils.validate_conductor_group('group1.with.dot'))
        self.assertIsNone(utils.validate_conductor_group('group1_with_under'))
        self.assertIsNone(utils.validate_conductor_group('group1-with-dash'))

    def test_validate_conductor_group_fail(self):
        self.assertRaises(exception.InvalidConductorGroup,
                          utils.validate_conductor_group, 'foo:bar')
        self.assertRaises(exception.InvalidConductorGroup,
                          utils.validate_conductor_group, 'foo*bar')
        self.assertRaises(exception.InvalidConductorGroup,
                          utils.validate_conductor_group, 'foo$bar')
        self.assertRaises(exception.InvalidConductorGroup,
                          utils.validate_conductor_group, object())
        self.assertRaises(exception.InvalidConductorGroup,
                          utils.validate_conductor_group, None)


class UnlinkTestCase(base.TestCase):

    def test_unlink(self):
        with mock.patch.object(os, "unlink", autospec=True) as unlink_mock:
            unlink_mock.return_value = None
            utils.unlink_without_raise("/fake/path")
            unlink_mock.assert_called_once_with("/fake/path")

    def test_unlink_ENOENT(self):
        with mock.patch.object(os, "unlink", autospec=True) as unlink_mock:
            unlink_mock.side_effect = OSError(errno.ENOENT)
            utils.unlink_without_raise("/fake/path")
            unlink_mock.assert_called_once_with("/fake/path")


class ExecuteTestCase(base.TestCase):
    # Allow calls to utils.execute() and related functions
    block_execute = False

    @mock.patch.object(processutils, 'execute', autospec=True)
    @mock.patch.object(os.environ, 'copy', return_value={}, autospec=True)
    def test_execute_use_standard_locale_no_env_variables(self, env_mock,
                                                          execute_mock):
        utils.execute('foo', use_standard_locale=True)
        execute_mock.assert_called_once_with('foo',
                                             env_variables={'LC_ALL': 'C'})

    @mock.patch.object(processutils, 'execute', autospec=True)
    def test_execute_use_standard_locale_with_env_variables(self,
                                                            execute_mock):
        utils.execute('foo', use_standard_locale=True,
                      env_variables={'foo': 'bar'})
        execute_mock.assert_called_once_with('foo',
                                             env_variables={'LC_ALL': 'C',
                                                            'foo': 'bar'})

    @mock.patch.object(processutils, 'execute', autospec=True)
    def test_execute_not_use_standard_locale(self, execute_mock):
        utils.execute('foo', use_standard_locale=False,
                      env_variables={'foo': 'bar'})
        execute_mock.assert_called_once_with('foo',
                                             env_variables={'foo': 'bar'})

    @mock.patch.object(utils, 'LOG', autospec=True)
    def _test_execute_with_log_stdout(self, log_mock, log_stdout=None):
        with mock.patch.object(
                processutils, 'execute', autospec=True) as execute_mock:
            execute_mock.return_value = ('stdout', 'stderr')
            if log_stdout is not None:
                utils.execute('foo', log_stdout=log_stdout)
            else:
                utils.execute('foo')
            execute_mock.assert_called_once_with('foo')
            name, args, kwargs = log_mock.debug.mock_calls[0]
            if log_stdout is False:
                self.assertEqual(1, log_mock.debug.call_count)
                self.assertNotIn('stdout', args[0])
            else:
                self.assertEqual(2, log_mock.debug.call_count)
                self.assertIn('stdout', args[0])

    def test_execute_with_log_stdout_default(self):
        self._test_execute_with_log_stdout()

    def test_execute_with_log_stdout_true(self):
        self._test_execute_with_log_stdout(log_stdout=True)

    def test_execute_with_log_stdout_false(self):
        self._test_execute_with_log_stdout(log_stdout=False)

    @mock.patch.object(utils, 'LOG', autospec=True)
    @mock.patch.object(processutils, 'execute', autospec=True)
    def test_execute_command_not_found(self, execute_mock, log_mock):
        execute_mock.side_effect = FileNotFoundError
        self.assertRaises(FileNotFoundError, utils.execute, 'foo')
        execute_mock.assert_called_once_with('foo')
        name, args, kwargs = log_mock.debug.mock_calls[0]
        self.assertEqual(1, log_mock.debug.call_count)
        self.assertIn('not found', args[0])


class MkfsTestCase(base.TestCase):

    @mock.patch.object(utils, 'execute', autospec=True)
    def test_mkfs(self, execute_mock):
        utils.mkfs('ext4', '/my/block/dev')
        utils.mkfs('msdos', '/my/msdos/block/dev')
        utils.mkfs('swap', '/my/swap/block/dev')

        expected = [mock.call('mkfs', '-t', 'ext4', '-F', '/my/block/dev',
                              use_standard_locale=True),
                    mock.call('mkfs', '-t', 'msdos', '/my/msdos/block/dev',
                              use_standard_locale=True),
                    mock.call('mkswap', '/my/swap/block/dev',
                              use_standard_locale=True)]
        self.assertEqual(expected, execute_mock.call_args_list)

    @mock.patch.object(utils, 'execute', autospec=True)
    def test_mkfs_with_label(self, execute_mock):
        utils.mkfs('ext4', '/my/block/dev', 'ext4-vol')
        utils.mkfs('msdos', '/my/msdos/block/dev', 'msdos-vol')
        utils.mkfs('swap', '/my/swap/block/dev', 'swap-vol')

        expected = [mock.call('mkfs', '-t', 'ext4', '-F', '-L', 'ext4-vol',
                              '/my/block/dev',
                              use_standard_locale=True),
                    mock.call('mkfs', '-t', 'msdos', '-n', 'msdos-vol',
                              '/my/msdos/block/dev',
                              use_standard_locale=True),
                    mock.call('mkswap', '-L', 'swap-vol',
                              '/my/swap/block/dev',
                              use_standard_locale=True)]
        self.assertEqual(expected, execute_mock.call_args_list)

    @mock.patch.object(utils, 'execute', autospec=True,
                       side_effect=processutils.ProcessExecutionError(
                           stderr=os.strerror(errno.ENOENT)))
    def test_mkfs_with_unsupported_fs(self, execute_mock):
        self.assertRaises(exception.FileSystemNotSupported,
                          utils.mkfs, 'foo', '/my/block/dev')

    @mock.patch.object(utils, 'execute', autospec=True,
                       side_effect=processutils.ProcessExecutionError(
                           stderr='fake'))
    def test_mkfs_with_unexpected_error(self, execute_mock):
        self.assertRaises(processutils.ProcessExecutionError, utils.mkfs,
                          'ext4', '/my/block/dev', 'ext4-vol')


class IsHttpUrlTestCase(base.TestCase):

    def test_is_http_url(self):
        self.assertTrue(utils.is_http_url('http://127.0.0.1'))
        self.assertTrue(utils.is_http_url('https://127.0.0.1'))
        self.assertTrue(utils.is_http_url('HTTP://127.1.2.3'))
        self.assertTrue(utils.is_http_url('HTTPS://127.3.2.1'))
        self.assertFalse(utils.is_http_url('Zm9vYmFy'))
        self.assertFalse(utils.is_http_url('11111111'))


class ParseRootDeviceTestCase(base.TestCase):

    def test_parse_root_device_hints_without_operators(self):
        root_device = {
            'wwn': '123456', 'model': 'FOO model', 'size': 12345,
            'serial': 'foo-serial', 'vendor': 'foo VENDOR with space',
            'name': '/dev/sda', 'wwn_with_extension': '123456111',
            'wwn_vendor_extension': '111', 'rotational': True,
            'hctl': '1:0:0:0', 'by_path': '/dev/disk/by-path/1:0:0:0'}
        result = utils.parse_root_device_hints(root_device)
        expected = {
            'wwn': 's== 123456', 'model': 's== foo%20model',
            'size': '== 12345', 'serial': 's== foo-serial',
            'vendor': 's== foo%20vendor%20with%20space',
            'name': 's== /dev/sda', 'wwn_with_extension': 's== 123456111',
            'wwn_vendor_extension': 's== 111', 'rotational': True,
            'hctl': 's== 1%3A0%3A0%3A0',
            'by_path': 's== /dev/disk/by-path/1%3A0%3A0%3A0'}
        self.assertEqual(expected, result)

    def test_parse_root_device_hints_with_operators(self):
        root_device = {
            'wwn': 's== 123456', 'model': 's== foo MODEL', 'size': '>= 12345',
            'serial': 's!= foo-serial', 'vendor': 's== foo VENDOR with space',
            'name': '<or> /dev/sda <or> /dev/sdb',
            'wwn_with_extension': 's!= 123456111',
            'wwn_vendor_extension': 's== 111', 'rotational': True,
            'hctl': 's== 1:0:0:0', 'by_path': 's== /dev/disk/by-path/1:0:0:0'}

        # Validate strings being normalized
        expected = copy.deepcopy(root_device)
        expected['model'] = 's== foo%20model'
        expected['vendor'] = 's== foo%20vendor%20with%20space'
        expected['hctl'] = 's== 1%3A0%3A0%3A0'
        expected['by_path'] = 's== /dev/disk/by-path/1%3A0%3A0%3A0'

        result = utils.parse_root_device_hints(root_device)
        # The hints already contain the operators, make sure we keep it
        self.assertEqual(expected, result)

    def test_parse_root_device_hints_string_compare_operator_name(self):
        root_device = {'name': 's== /dev/sdb'}
        # Validate strings being normalized
        expected = copy.deepcopy(root_device)
        result = utils.parse_root_device_hints(root_device)
        # The hints already contain the operators, make sure we keep it
        self.assertEqual(expected, result)

    def test_parse_root_device_hints_no_hints(self):
        result = utils.parse_root_device_hints({})
        self.assertIsNone(result)

    def test_parse_root_device_hints_convert_size(self):
        for size in (12345, '12345'):
            result = utils.parse_root_device_hints({'size': size})
            self.assertEqual({'size': '== 12345'}, result)

    def test_parse_root_device_hints_invalid_size(self):
        for value in ('not-int', -123, 0):
            self.assertRaises(ValueError, utils.parse_root_device_hints,
                              {'size': value})

    def test_parse_root_device_hints_int_or(self):
        expr = '<or> 123 <or> 456 <or> 789'
        result = utils.parse_root_device_hints({'size': expr})
        self.assertEqual({'size': expr}, result)

    def test_parse_root_device_hints_int_or_invalid(self):
        expr = '<or> 123 <or> non-int <or> 789'
        self.assertRaises(ValueError, utils.parse_root_device_hints,
                          {'size': expr})

    def test_parse_root_device_hints_string_or_space(self):
        expr = '<or> foo <or> foo bar <or> bar'
        expected = '<or> foo <or> foo%20bar <or> bar'
        result = utils.parse_root_device_hints({'model': expr})
        self.assertEqual({'model': expected}, result)

    def _parse_root_device_hints_convert_rotational(self, values,
                                                    expected_value):
        for value in values:
            result = utils.parse_root_device_hints({'rotational': value})
            self.assertEqual({'rotational': expected_value}, result)

    def test_parse_root_device_hints_convert_rotational(self):
        self._parse_root_device_hints_convert_rotational(
            (True, 'true', 'on', 'y', 'yes'), True)

        self._parse_root_device_hints_convert_rotational(
            (False, 'false', 'off', 'n', 'no'), False)

    def test_parse_root_device_hints_invalid_rotational(self):
        self.assertRaises(ValueError, utils.parse_root_device_hints,
                          {'rotational': 'not-bool'})

    def test_parse_root_device_hints_invalid_wwn(self):
        self.assertRaises(ValueError, utils.parse_root_device_hints,
                          {'wwn': 123})

    def test_parse_root_device_hints_invalid_wwn_with_extension(self):
        self.assertRaises(ValueError, utils.parse_root_device_hints,
                          {'wwn_with_extension': 123})

    def test_parse_root_device_hints_invalid_wwn_vendor_extension(self):
        self.assertRaises(ValueError, utils.parse_root_device_hints,
                          {'wwn_vendor_extension': 123})

    def test_parse_root_device_hints_invalid_model(self):
        self.assertRaises(ValueError, utils.parse_root_device_hints,
                          {'model': 123})

    def test_parse_root_device_hints_invalid_serial(self):
        self.assertRaises(ValueError, utils.parse_root_device_hints,
                          {'serial': 123})

    def test_parse_root_device_hints_invalid_vendor(self):
        self.assertRaises(ValueError, utils.parse_root_device_hints,
                          {'vendor': 123})

    def test_parse_root_device_hints_invalid_name(self):
        self.assertRaises(ValueError, utils.parse_root_device_hints,
                          {'name': 123})

    def test_parse_root_device_hints_invalid_hctl(self):
        self.assertRaises(ValueError, utils.parse_root_device_hints,
                          {'hctl': 123})

    def test_parse_root_device_hints_invalid_by_path(self):
        self.assertRaises(ValueError, utils.parse_root_device_hints,
                          {'by_path': 123})

    def test_parse_root_device_hints_non_existent_hint(self):
        self.assertRaises(ValueError, utils.parse_root_device_hints,
                          {'non-existent': 'foo'})

    def test_extract_hint_operator_and_values_single_value(self):
        expected = {'op': '>=', 'values': ['123']}
        self.assertEqual(
            expected, utils._extract_hint_operator_and_values(
                '>= 123', 'size'))

    def test_extract_hint_operator_and_values_multiple_values(self):
        expected = {'op': '<or>', 'values': ['123', '456', '789']}
        expr = '<or> 123 <or> 456 <or> 789'
        self.assertEqual(
            expected, utils._extract_hint_operator_and_values(expr, 'size'))

    def test_extract_hint_operator_and_values_multiple_values_space(self):
        expected = {'op': '<or>', 'values': ['foo', 'foo bar', 'bar']}
        expr = '<or> foo <or> foo bar <or> bar'
        self.assertEqual(
            expected, utils._extract_hint_operator_and_values(expr, 'model'))

    def test_extract_hint_operator_and_values_no_operator(self):
        expected = {'op': '', 'values': ['123']}
        self.assertEqual(
            expected, utils._extract_hint_operator_and_values('123', 'size'))

    def test_extract_hint_operator_and_values_empty_value(self):
        self.assertRaises(
            ValueError, utils._extract_hint_operator_and_values, '', 'size')

    def test_extract_hint_operator_and_values_integer(self):
        expected = {'op': '', 'values': ['123']}
        self.assertEqual(
            expected, utils._extract_hint_operator_and_values(123, 'size'))

    def test__append_operator_to_hints(self):
        root_device = {'serial': 'foo', 'size': 12345,
                       'model': 'foo model', 'rotational': True}
        expected = {'serial': 's== foo', 'size': '== 12345',
                    'model': 's== foo model', 'rotational': True}

        result = utils._append_operator_to_hints(root_device)
        self.assertEqual(expected, result)

    def test_normalize_hint_expression_or(self):
        expr = '<or> foo <or> foo bar <or> bar'
        expected = '<or> foo <or> foo%20bar <or> bar'
        result = utils._normalize_hint_expression(expr, 'model')
        self.assertEqual(expected, result)

    def test_normalize_hint_expression_in(self):
        expr = '<in> foo <in> foo bar <in> bar'
        expected = '<in> foo <in> foo%20bar <in> bar'
        result = utils._normalize_hint_expression(expr, 'model')
        self.assertEqual(expected, result)

    def test_normalize_hint_expression_op_space(self):
        expr = 's== test string with space'
        expected = 's== test%20string%20with%20space'
        result = utils._normalize_hint_expression(expr, 'model')
        self.assertEqual(expected, result)

    def test_normalize_hint_expression_op_no_space(self):
        expr = 's!= SpongeBob'
        expected = 's!= spongebob'
        result = utils._normalize_hint_expression(expr, 'model')
        self.assertEqual(expected, result)

    def test_normalize_hint_expression_no_op_space(self):
        expr = 'no operators'
        expected = 'no%20operators'
        result = utils._normalize_hint_expression(expr, 'model')
        self.assertEqual(expected, result)

    def test_normalize_hint_expression_no_op_no_space(self):
        expr = 'NoSpace'
        expected = 'nospace'
        result = utils._normalize_hint_expression(expr, 'model')
        self.assertEqual(expected, result)

    def test_normalize_hint_expression_empty_value(self):
        self.assertRaises(
            ValueError, utils._normalize_hint_expression, '', 'size')


class MatchRootDeviceTestCase(base.TestCase):

    def setUp(self):
        super(MatchRootDeviceTestCase, self).setUp()
        self.devices = [
            {'name': '/dev/sda', 'size': 64424509440, 'model': 'ok model',
             'serial': 'fakeserial'},
            {'name': '/dev/sdb', 'size': 128849018880, 'model': 'big model',
             'serial': 'veryfakeserial', 'rotational': 'yes'},
            {'name': '/dev/sdc', 'size': 10737418240, 'model': 'small model',
             'serial': 'veryveryfakeserial', 'rotational': False},
        ]

    def test_match_root_device_hints_one_hint(self):
        root_device_hints = {'size': '>= 70'}
        dev = utils.match_root_device_hints(self.devices, root_device_hints)
        self.assertEqual('/dev/sdb', dev['name'])

    def test_match_root_device_hints_rotational(self):
        root_device_hints = {'rotational': False}
        dev = utils.match_root_device_hints(self.devices, root_device_hints)
        self.assertEqual('/dev/sdc', dev['name'])

    def test_match_root_device_hints_rotational_convert_devices_bool(self):
        root_device_hints = {'size': '>=100', 'rotational': True}
        dev = utils.match_root_device_hints(self.devices, root_device_hints)
        self.assertEqual('/dev/sdb', dev['name'])

    def test_match_root_device_hints_multiple_hints(self):
        root_device_hints = {'size': '>= 50', 'model': 's==big model',
                             'serial': 's==veryfakeserial'}
        dev = utils.match_root_device_hints(self.devices, root_device_hints)
        self.assertEqual('/dev/sdb', dev['name'])

    def test_match_root_device_hints_multiple_hints2(self):
        root_device_hints = {
            'size': '<= 20',
            'model': '<or> model 5 <or> foomodel <or> small model <or>',
            'serial': 's== veryveryfakeserial'}
        dev = utils.match_root_device_hints(self.devices, root_device_hints)
        self.assertEqual('/dev/sdc', dev['name'])

    def test_match_root_device_hints_multiple_hints3(self):
        root_device_hints = {'rotational': False, 'model': '<in> small'}
        dev = utils.match_root_device_hints(self.devices, root_device_hints)
        self.assertEqual('/dev/sdc', dev['name'])

    def test_match_root_device_hints_no_operators(self):
        root_device_hints = {'size': '120', 'model': 'big model',
                             'serial': 'veryfakeserial'}
        dev = utils.match_root_device_hints(self.devices, root_device_hints)
        self.assertEqual('/dev/sdb', dev['name'])

    def test_match_root_device_hints_no_device_found(self):
        root_device_hints = {'size': '>=50', 'model': 's==foo'}
        dev = utils.match_root_device_hints(self.devices, root_device_hints)
        self.assertIsNone(dev)

    @mock.patch.object(utils.LOG, 'warning', autospec=True)
    def test_match_root_device_hints_empty_device_attribute(self, mock_warn):
        empty_dev = [{'name': '/dev/sda', 'model': ' '}]
        dev = utils.match_root_device_hints(empty_dev, {'model': 'foo'})
        self.assertIsNone(dev)
        self.assertTrue(mock_warn.called)

    def test_find_devices_all(self):
        root_device_hints = {'size': '>= 10'}
        devs = list(utils.find_devices_by_hints(self.devices,
                                                root_device_hints))
        self.assertEqual(self.devices, devs)

    def test_find_devices_none(self):
        root_device_hints = {'size': '>= 100500'}
        devs = list(utils.find_devices_by_hints(self.devices,
                                                root_device_hints))
        self.assertEqual([], devs)

    def test_find_devices_name(self):
        root_device_hints = {'name': 's== /dev/sda'}
        devs = list(utils.find_devices_by_hints(self.devices,
                                                root_device_hints))
        self.assertEqual([self.devices[0]], devs)


@mock.patch.object(utils, 'execute', autospec=True)
class GetRouteSourceTestCase(base.TestCase):

    def test_get_route_source_ipv4(self, mock_execute):
        mock_execute.return_value = ('XXX src 1.2.3.4 XXX\n    cache', None)

        source = utils.get_route_source('XXX')
        self.assertEqual('1.2.3.4', source)

    def test_get_route_source_ipv6(self, mock_execute):
        mock_execute.return_value = ('XXX src 1:2::3:4 metric XXX\n    cache',
                                     None)

        source = utils.get_route_source('XXX')
        self.assertEqual('1:2::3:4', source)

    def test_get_route_source_ipv6_linklocal(self, mock_execute):
        mock_execute.return_value = (
            'XXX src fe80::1234:1234:1234:1234 metric XXX\n    cache', None)

        source = utils.get_route_source('XXX')
        self.assertIsNone(source)

    def test_get_route_source_ipv6_linklocal_allowed(self, mock_execute):
        mock_execute.return_value = (
            'XXX src fe80::1234:1234:1234:1234 metric XXX\n    cache', None)

        source = utils.get_route_source('XXX', ignore_link_local=False)
        self.assertEqual('fe80::1234:1234:1234:1234', source)

    def test_get_route_source_indexerror(self, mock_execute):
        mock_execute.return_value = ('XXX src \n    cache', None)

        source = utils.get_route_source('XXX')
        self.assertIsNone(source)



# Copyright 2018 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Test class for Raid Interface used by iLO5."""

from unittest import mock

from oslo_utils import importutils

from ironic.common import exception
from ironic.common import raid
from ironic.common import states
from ironic.conductor import task_manager
from ironic.conductor import utils as manager_utils
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules.ilo import common as ilo_common
from ironic.drivers.modules.ilo import raid as ilo_raid
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.db import utils as db_utils
from ironic.tests.unit.objects import utils as obj_utils

ilo_error = importutils.try_import('proliantutils.exception')

INFO_DICT = db_utils.get_test_ilo_info()


class Ilo5RAIDTestCase(db_base.DbTestCase):

    def setUp(self):
        super(Ilo5RAIDTestCase, self).setUp()
        self.driver = mock.Mock(raid=ilo_raid.Ilo5RAID())
        self.target_raid_config = {
            "logical_disks": [
                {'size_gb': 200, 'raid_level': 0, 'is_root_volume': True},
                {'size_gb': 200, 'raid_level': 5}
            ]}
        n = {
            'driver': 'ilo5',
            'driver_info': INFO_DICT,
            'target_raid_config': self.target_raid_config,
        }
        self.config(enabled_hardware_types=['ilo5'],
                    enabled_boot_interfaces=['ilo-virtual-media'],
                    enabled_console_interfaces=['ilo'],
                    enabled_deploy_interfaces=['direct'],
                    enabled_inspect_interfaces=['ilo'],
                    enabled_management_interfaces=['ilo5'],
                    enabled_power_interfaces=['ilo'],
                    enabled_raid_interfaces=['ilo5'])
        self.node = obj_utils.create_test_node(self.context, **n)

    @mock.patch.object(deploy_utils, 'build_agent_options', autospec=True)
    @mock.patch.object(manager_utils, 'node_power_action', autospec=True)
    def _test__prepare_for_read_raid_create_raid(
            self, mock_reboot, mock_build_opt):
        with task_manager.acquire(self.context, self.node.uuid) as task:
            mock_build_opt.return_value = []
            task.driver.raid._prepare_for_read_raid(task, 'create_raid')
            self.assertTrue(
                task.node.driver_internal_info.get(
                    'ilo_raid_create_in_progress'))
            if task.node.clean_step:
                self.assertTrue(
                    task.node.driver_internal_info.get(
                        'cleaning_reboot'))
                self.assertFalse(
                    task.node.driver_internal_info.get(
                        'skip_current_clean_step'))
            if task.node.deploy_step:
                self.assertTrue(
                    task.node.driver_internal_info.get(
                        'deployment_reboot'))
                self.assertFalse(
                    task.node.driver_internal_info.get(
                        'skip_current_deploy_step'))
            mock_reboot.assert_called_once_with(task, states.REBOOT, None)

    def test__prepare_for_read_raid_create_raid_cleaning(self):
        self.node.clean_step = {'step': 'create_configuration',
                                'interface': 'raid'}
        self.node.save()
        self._test__prepare_for_read_raid_create_raid()

    def test__prepare_for_read_raid_create_raid_deploying(self):
        self.node.deploy_step = {'step': 'create_configuration',
                                 'interface': 'raid'}
        self.node.save()
        self._test__prepare_for_read_raid_create_raid()

    @mock.patch.object(deploy_utils, 'build_agent_options', autospec=True)
    @mock.patch.object(manager_utils, 'node_power_action', autospec=True)
    def _test__prepare_for_read_raid_delete_raid(
            self, mock_reboot, mock_build_opt):
        with task_manager.acquire(self.context, self.node.uuid) as task:
            mock_build_opt.return_value = []
            task.driver.raid._prepare_for_read_raid(task, 'delete_raid')
            self.assertTrue(
                task.node.driver_internal_info.get(
                    'ilo_raid_delete_in_progress'))
            if task.node.clean_step:
                self.assertTrue(
                    task.node.driver_internal_info.get(
                        'cleaning_reboot'))
                self.assertEqual(
                    task.node.driver_internal_info.get(
                        'skip_current_clean_step'), False)
            else:
                self.assertTrue(
                    task.node.driver_internal_info.get(
                        'deployment_reboot'))
                self.assertEqual(
                    task.node.driver_internal_info.get(
                        'skip_current_deploy_step'), False)
            mock_reboot.assert_called_once_with(task, states.REBOOT, None)

    def test__prepare_for_read_raid_delete_raid_cleaning(self):
        self.node.clean_step = {'step': 'create_configuration',
                                'interface': 'raid'}
        self.node.save()
        self._test__prepare_for_read_raid_delete_raid()

    def test__prepare_for_read_raid_delete_raid_deploying(self):
        self.node.deploy_step = {'step': 'create_configuration',
                                 'interface': 'raid'}
        self.node.save()
        self._test__prepare_for_read_raid_delete_raid()

    @mock.patch.object(ilo_raid.Ilo5RAID, '_prepare_for_read_raid',
                       autospec=True)
    @mock.patch.object(raid, 'filter_target_raid_config', autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    def _test_create_configuration(
            self, ilo_mock, filter_target_raid_config_mock, prepare_raid_mock):
        ilo_mock_object = ilo_mock.return_value
        with task_manager.acquire(self.context, self.node.uuid) as task:
            filter_target_raid_config_mock.return_value = (
                self.target_raid_config)
            result = task.driver.raid.create_configuration(task)
            prepare_raid_mock.assert_called_once_with(
                mock.ANY, task, 'create_raid')
            if task.node.clean_step:
                self.assertEqual(states.CLEANWAIT, result)
            else:
                self.assertEqual(states.DEPLOYWAIT, result)
        (ilo_mock_object.create_raid_configuration.
         assert_called_once_with(self.target_raid_config))

    def test_create_configuration_cleaning(self):
        self.node.clean_step = {'step': 'create_configuration',
                                'interface': 'raid'}
        self.node.save()
        self._test_create_configuration()

    def test_create_configuration_deploying(self):
        self.node.deploy_step = {'step': 'create_configuration',
                                 'interface': 'raid'}
        self.node.save()
        self._test_create_configuration()

    @mock.patch.object(raid, 'update_raid_info', autospec=True)
    @mock.patch.object(raid, 'filter_target_raid_config', autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    def _test_create_configuration_with_read_raid(
            self, ilo_mock, filter_target_raid_config_mock, update_raid_mock):
        raid_conf = {u'logical_disks':
                     [{u'size_gb': 89,
                       u'physical_disks': [u'5I:1:1'],
                       u'raid_level': u'0',
                       u'root_device_hint': {u'wwn': u'0x600508b1001c7e87'},
                       u'controller': u'Smart Array P822 in Slot 1',
                       u'volume_name': u'0006EB7BPDVTF0BRH5L0EAEDDA'}]
                     }
        ilo_mock_object = ilo_mock.return_value
        driver_internal_info = self.node.driver_internal_info
        driver_internal_info['ilo_raid_create_in_progress'] = True
        if self.node.clean_step:
            driver_internal_info['skip_current_clean_step'] = False
            driver_internal_info['cleaning_reboot'] = True
        else:
            driver_internal_info['skip_current_deploy_step'] = False
            driver_internal_info['deployment_reboot'] = True
        self.node.driver_internal_info = driver_internal_info
        self.node.save()

        with task_manager.acquire(self.context, self.node.uuid) as task:
            filter_target_raid_config_mock.return_value = (
                self.target_raid_config)
            ilo_mock_object.read_raid_configuration.return_value = raid_conf
            task.driver.raid.create_configuration(task)
            update_raid_mock.assert_called_once_with(task.node, raid_conf)
            self.assertNotIn('ilo_raid_create_in_progress',
                             task.node.driver_internal_info)
            if task.node.clean_step:
                self.assertNotIn('skip_current_clean_step',
                                 task.node.driver_internal_info)

    def test_create_configuration_with_read_raid_cleaning(self):
        self.node.clean_step = {'step': 'create_configuration',
                                'interface': 'raid'}
        self.node.save()
        self._test_create_configuration_with_read_raid()

    def test_create_configuration_with_read_raid_deploying(self):
        self.node.deploy_step = {'step': 'create_configuration',
                                 'interface': 'raid'}
        self.node.save()
        self._test_create_configuration_with_read_raid()

    @mock.patch.object(raid, 'filter_target_raid_config', autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    def _test_create_configuration_with_read_raid_failed(
            self, ilo_mock, filter_target_raid_config_mock):
        raid_conf = {u'logical_disks': []}
        driver_internal_info = self.node.driver_internal_info
        driver_internal_info['ilo_raid_create_in_progress'] = True
        driver_internal_info['skip_current_clean_step'] = False
        self.node.driver_internal_info = driver_internal_info
        self.node.save()
        ilo_mock_object = ilo_mock.return_value
        if self.node.clean_step:
            exept = exception.NodeCleaningFailure
        else:
            exept = exception.InstanceDeployFailure
        with task_manager.acquire(self.context, self.node.uuid) as task:
            filter_target_raid_config_mock.return_value = (
                self.target_raid_config)
            ilo_mock_object.read_raid_configuration.return_value = raid_conf
            self.assertRaises(exept,
                              task.driver.raid.create_configuration, task)
            self.assertNotIn('ilo_raid_create_in_progress',
                             task.node.driver_internal_info)
            if task.node.clean_step:
                self.assertNotIn('skip_current_clean_step',
                                 task.node.driver_internal_info)
            else:
                self.assertNotIn('skip_current_deploy_step',
                                 task.node.driver_internal_info)

    def test_create_configuration_with_read_raid_failed_cleaning(self):
        self.node.clean_step = {'step': 'create_configuration',
                                'interface': 'raid'}
        self.node.save()
        self._test_create_configuration_with_read_raid_failed()

    def test_create_configuration_with_read_raid_failed_deploying(self):
        self.node.deploy_step = {'step': 'create_configuration',
                                 'interface': 'raid'}
        self.node.save()
        self._test_create_configuration_with_read_raid_failed()

    @mock.patch.object(raid, 'filter_target_raid_config', autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    def _test_create_configuration_empty_target_raid_config(
            self, ilo_mock, filter_target_raid_config_mock):
        self.node.target_raid_config = {}
        self.node.save()
        ilo_mock_object = ilo_mock.return_value
        with task_manager.acquire(self.context, self.node.uuid) as task:
            msg = "Node %s has no target RAID configuration" % self.node.uuid
            filter_target_raid_config_mock.side_effect = (
                exception.MissingParameterValue(msg))
            self.assertRaises(exception.MissingParameterValue,
                              task.driver.raid.create_configuration, task)
        self.assertFalse(ilo_mock_object.create_raid_configuration.called)

    def test_create_configuration_empty_target_raid_config_cleaning(self):
        self.node.clean_step = {'step': 'create_configuration',
                                'interface': 'raid'}
        self.node.save()
        self._test_create_configuration_empty_target_raid_config()

    def test_create_configuration_empty_target_raid_config_deploying(self):
        self.node.deploy_step = {'step': 'create_configuration',
                                 'interface': 'raid'}
        self.node.save()
        self._test_create_configuration_empty_target_raid_config()

    @mock.patch.object(ilo_raid.Ilo5RAID, '_prepare_for_read_raid',
                       autospec=True)
    @mock.patch.object(raid, 'filter_target_raid_config', autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    def _test_create_configuration_skip_root(
            self, ilo_mock, filter_target_raid_config_mock,
            prepare_raid_mock):
        ilo_mock_object = ilo_mock.return_value
        with task_manager.acquire(self.context, self.node.uuid) as task:
            exp_target_raid_config = {
                "logical_disks": [
                    {'size_gb': 200, 'raid_level': 5}
                ]}
            filter_target_raid_config_mock.return_value = (
                exp_target_raid_config)
            result = task.driver.raid.create_configuration(
                task, create_root_volume=False)
            (ilo_mock_object.create_raid_configuration.
             assert_called_once_with(exp_target_raid_config))
            if task.node.clean_step:
                self.assertEqual(states.CLEANWAIT, result)
            else:
                self.assertEqual(states.DEPLOYWAIT, result)
            prepare_raid_mock.assert_called_once_with(
                mock.ANY, task, 'create_raid')
            self.assertEqual(
                exp_target_raid_config,
                task.node.driver_internal_info['target_raid_config'])

    def test_create_configuration_skip_root_cleaning(self):
        self.node.clean_step = {'step': 'create_configuration',
                                'interface': 'raid'}
        self.node.save()
        self._test_create_configuration_skip_root()

    def test_create_configuration_skip_root_deploying(self):
        self.node.deploy_step = {'step': 'create_configuration',
                                 'interface': 'raid'}
        self.node.save()
        self._test_create_configuration_skip_root()

    @mock.patch.object(ilo_raid.Ilo5RAID, '_prepare_for_read_raid',
                       autospec=True)
    @mock.patch.object(raid, 'filter_target_raid_config', autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    def _test_create_configuration_skip_non_root(
            self, ilo_mock, filter_target_raid_config_mock, prepare_raid_mock):
        ilo_mock_object = ilo_mock.return_value
        with task_manager.acquire(self.context, self.node.uuid) as task:
            exp_target_raid_config = {
                "logical_disks": [
                    {'size_gb': 200, 'raid_level': 0, 'is_root_volume': True}
                ]}
            filter_target_raid_config_mock.return_value = (
                exp_target_raid_config)
            result = task.driver.raid.create_configuration(
                task, create_nonroot_volumes=False)
            (ilo_mock_object.create_raid_configuration.
             assert_called_once_with(exp_target_raid_config))
            prepare_raid_mock.assert_called_once_with(
                mock.ANY, task, 'create_raid')
            if task.node.clean_step:
                self.assertEqual(states.CLEANWAIT, result)
            else:
                self.assertEqual(states.DEPLOYWAIT, result)
            self.assertEqual(
                exp_target_raid_config,
                task.node.driver_internal_info['target_raid_config'])

    def test_create_configuration_skip_non_root_cleaning(self):
        self.node.clean_step = {'step': 'create_configuration',
                                'interface': 'raid'}
        self.node.save()
        self._test_create_configuration_skip_non_root()

    def test_create_configuration_skip_non_root_deploying(self):
        self.node.deploy_step = {'step': 'create_configuration',
                                 'interface': 'raid'}
        self.node.save()
        self._test_create_configuration_skip_non_root()

    @mock.patch.object(raid, 'filter_target_raid_config', autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    def _test_create_configuration_skip_root_skip_non_root(
            self, ilo_mock, filter_target_raid_config_mock):
        ilo_mock_object = ilo_mock.return_value
        with task_manager.acquire(self.context, self.node.uuid) as task:
            msg = "Node %s has no target RAID configuration" % self.node.uuid
            filter_target_raid_config_mock.side_effect = (
                exception.MissingParameterValue(msg))
            self.assertRaises(
                exception.MissingParameterValue,
                task.driver.raid.create_configuration,
                task, False, False)
            self.assertFalse(ilo_mock_object.create_raid_configuration.called)

    def test_create_configuration_skip_root_skip_non_root_cleaning(self):
        self.node.clean_step = {'step': 'create_configuration',
                                'interface': 'raid'}
        self.node.save()
        self._test_create_configuration_skip_root_skip_non_root()

    def test_create_configuration_skip_root_skip_non_root_deploying(self):
        self.node.deploy_step = {'step': 'create_configuration',
                                 'interface': 'raid'}
        self.node.save()
        self._test_create_configuration_skip_root_skip_non_root()

    @mock.patch.object(manager_utils, 'cleaning_error_handler',
                       autospec=True)
    @mock.patch.object(manager_utils, 'deploying_error_handler',
                       autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    def _test_create_configuration_ilo_error(self, ilo_mock,
                                             deploy_err_handler_mock,
                                             clean_err_handler_mock):
        ilo_mock_object = ilo_mock.return_value
        exc = ilo_error.IloError('error')
        ilo_mock_object.create_raid_configuration.side_effect = exc
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.driver.raid.create_configuration(
                task, create_nonroot_volumes=False)
            self.assertNotIn('ilo_raid_create_in_progress',
                             task.node.driver_internal_info)
            if task.node.clean_step:
                self.assertNotIn('skip_current_clean_step',
                                 task.node.driver_internal_info)
                clean_err_handler_mock.assert_called_once_with(
                    task,
                    ("RAID configuration job failed for node %s. Message: "
                     "'Failed to create raid configuration on node %s'." %
                     (self.node.uuid, self.node.uuid)),
                    'Failed to create raid configuration '
                    'on node %s' % self.node.uuid)
            else:
                self.assertNotIn('skip_current_deploy_step',
                                 task.node.driver_internal_info)
                deploy_err_handler_mock.assert_called_once_with(
                    task,
                    ("RAID configuration job failed for node %s. Message: "
                     "'Failed to create raid configuration on node %s'." %
                     (self.node.uuid, self.node.uuid)),
                    'Failed to create raid configuration '
                    'on node %s' % self.node.uuid)

    def test_create_configuration_ilo_error_cleaning(self):
        self.node.clean_step = {'step': 'create_configuration',
                                'interface': 'raid'}
        self.node.provision_state = states.CLEANING
        self.node.save()
        self._test_create_configuration_ilo_error()

    def test_create_configuration_ilo_error_cleaning_deploying(self):
        self.node.deploy_step = {'step': 'create_configuration',
                                 'interface': 'raid'}
        self.node.provision_state = states.DEPLOYING
        self.node.save()
        self._test_create_configuration_ilo_error()

    @mock.patch.object(ilo_raid.Ilo5RAID, '_prepare_for_read_raid',
                       autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    def _test_delete_configuration(self, ilo_mock, prepare_raid_mock):
        ilo_mock_object = ilo_mock.return_value
        with task_manager.acquire(self.context, self.node.uuid) as task:
            result = task.driver.raid.delete_configuration(task)
            if task.node.clean_step:
                self.assertEqual(states.CLEANWAIT, result)
            else:
                self.assertEqual(states.DEPLOYWAIT, result)
        ilo_mock_object.delete_raid_configuration.assert_called_once_with()
        prepare_raid_mock.assert_called_once_with(mock.ANY, task,
                                                  'delete_raid')

    def test_delete_configuration_cleaning(self):
        self.node.clean_step = {'step': 'create_configuration',
                                'interface': 'raid'}
        self.node.save()
        self._test_delete_configuration()

    def test_delete_configuration_deploying(self):
        self.node.deploy_step = {'step': 'create_configuration',
                                 'interface': 'raid'}
        self.node.save()
        self._test_delete_configuration()

    @mock.patch.object(ilo_raid.LOG, 'info', spec_set=True,
                       autospec=True)
    @mock.patch.object(ilo_raid.Ilo5RAID, '_prepare_for_read_raid',
                       autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    def _test_delete_configuration_no_logical_drive(
            self, ilo_mock, prepare_raid_mock, log_mock):
        ilo_mock_object = ilo_mock.return_value
        exc = ilo_error.IloLogicalDriveNotFoundError('No logical drive found')
        with task_manager.acquire(self.context, self.node.uuid) as task:
            ilo_mock_object.delete_raid_configuration.side_effect = exc
            task.driver.raid.delete_configuration(task)
            self.assertTrue(log_mock.called)

    def test_delete_configuration_no_logical_drive_cleaning(self):
        self.node.clean_step = {'step': 'create_configuration',
                                'interface': 'raid'}
        self.node.save()
        self._test_delete_configuration_no_logical_drive()

    def test_delete_configuration_no_logical_drive_deploying(self):
        self.node.deploy_step = {'step': 'create_configuration',
                                 'interface': 'raid'}
        self.node.save()
        self._test_delete_configuration_no_logical_drive()

    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    def _test_delete_configuration_with_read_raid(self, ilo_mock):
        raid_conf = {u'logical_disks': []}
        driver_internal_info = self.node.driver_internal_info
        driver_internal_info['ilo_raid_delete_in_progress'] = True
        driver_internal_info['skip_current_clean_step'] = False
        self.node.driver_internal_info = driver_internal_info
        self.node.save()
        ilo_mock_object = ilo_mock.return_value
        if self.node.clean_step:
            skip_field_name = 'skip_current_clean_step'
        else:
            skip_field_name = 'skip_current_deploy_step'
        with task_manager.acquire(self.context, self.node.uuid) as task:
            ilo_mock_object.read_raid_configuration.return_value = raid_conf
            task.driver.raid.delete_configuration(task)
            self.assertEqual(self.node.raid_config, {})
            self.assertNotIn('ilo_raid_delete_in_progress',
                             task.node.driver_internal_info)
            self.assertNotIn(skip_field_name,
                             task.node.driver_internal_info)

    def test_delete_configuration_with_read_raid_cleaning(self):
        self.node.clean_step = {'step': 'create_configuration',
                                'interface': 'raid'}
        self.node.save()
        self._test_delete_configuration_with_read_raid()

    def test_delete_configuration_with_read_raid_deploying(self):
        self.node.deploy_step = {'step': 'create_configuration',
                                 'interface': 'raid'}
        self.node.save()
        self._test_delete_configuration_with_read_raid()

    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    def _test_delete_configuration_with_read_raid_failed(self, ilo_mock):
        raid_conf = {u'logical_disks': [{'size_gb': 200,
                                         'raid_level': 0,
                                         'is_root_volume': True}]}
        driver_internal_info = self.node.driver_internal_info
        driver_internal_info['ilo_raid_delete_in_progress'] = True
        driver_internal_info['skip_current_clean_step'] = False
        self.node.driver_internal_info = driver_internal_info
        self.node.save()
        ilo_mock_object = ilo_mock.return_value
        if self.node.clean_step:
            exept = exception.NodeCleaningFailure
        else:
            exept = exception.InstanceDeployFailure
        with task_manager.acquire(self.context, self.node.uuid) as task:
            ilo_mock_object.read_raid_configuration.return_value = raid_conf
            self.assertRaises(exept,
                              task.driver.raid.delete_configuration, task)
            self.assertNotIn('ilo_raid_delete_in_progress',
                             task.node.driver_internal_info)
            if task.node.clean_step:
                self.assertNotIn('skip_current_clean_step',
                                 task.node.driver_internal_info)
            else:
                self.assertNotIn('skip_current_deploy_step',
                                 task.node.driver_internal_info)

    def test_delete_configuration_with_read_raid_failed_cleaning(self):
        self.node.clean_step = {'step': 'create_configuration',
                                'interface': 'raid'}
        self.node.save()
        self._test_delete_configuration_with_read_raid_failed()

    def test_delete_configuration_with_read_raid_failed_deploying(self):
        self.node.deploy_step = {'step': 'create_configuration',
                                 'interface': 'raid'}
        self.node.save()
        self._test_delete_configuration_with_read_raid_failed()

    @mock.patch.object(ilo_raid.Ilo5RAID, '_set_step_failed', autospec=True)
    @mock.patch.object(ilo_common, 'get_ilo_object', autospec=True)
    def _test_delete_configuration_ilo_error(self, ilo_mock,
                                             set_step_failed_mock):
        ilo_mock_object = ilo_mock.return_value
        exc = ilo_error.IloError('error')
        ilo_mock_object.delete_raid_configuration.side_effect = exc
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.driver.raid.delete_configuration(task)
            ilo_mock_object.delete_raid_configuration.assert_called_once_with()
            self.assertNotIn('ilo_raid_delete_in_progress',
                             task.node.driver_internal_info)
            self.assertNotIn('cleaning_reboot',
                             task.node.driver_internal_info)
            self.assertNotIn('skip_current_clean_step',
                             task.node.driver_internal_info)
            set_step_failed_mock.assert_called_once_with(
                mock.ANY,
                task,
                'Failed to delete raid configuration '
                'on node %s' % self.node.uuid, exc)

    def test_delete_configuration_ilo_error_cleaning(self):
        self.node.clean_step = {'step': 'create_configuration',
                                'interface': 'raid'}
        self.node.save()
        self._test_delete_configuration_ilo_error()

    def test_delete_configuration_ilo_error_deploying(self):
        self.node.deploy_step = {'step': 'create_configuration',
                                 'interface': 'raid'}
        self.node.save()
        self._test_delete_configuration_ilo_error()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging
from oslo_utils import netutils

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import utils
from ironic.conf import CONF
from ironic.drivers.modules.inspector.hooks import base

LOG = logging.getLogger(__name__)


class ValidateInterfacesHook(base.InspectionHook):
    """Hook to validate network interfaces."""

    def preprocess(self, task, inventory, plugin_data):
        all_interfaces = get_interfaces(task.node, inventory)
        valid_interfaces = validate_interfaces(task.node, inventory,
                                               all_interfaces)
        valid_macs = [iface['mac_address'] for iface in
                      valid_interfaces.values()]

        plugin_data['all_interfaces'] = all_interfaces
        plugin_data['valid_interfaces'] = valid_interfaces
        plugin_data['macs'] = valid_macs

    def __call__(self, task, inventory, plugin_data):
        pass


def get_interfaces(node, inventory):
    """Convert inventory to a dict with interfaces.

    :return: dict interface name -> interface (for valid interfaces).
    """
    result = {}
    pxe_mac = get_pxe_mac(inventory)

    for iface in inventory['interfaces']:
        name = iface.get('name')
        mac = iface.get('mac_address')
        ipv4_address = iface.get('ipv4_address')
        ipv6_address = iface.get('ipv6_address')
        # NOTE(kaifeng) ipv6 address may in the form of fd00::1%enp2s0,
        # which is not supported by netaddr, remove the suffix if exists.
        if ipv6_address and '%' in ipv6_address:
            ipv6_address = ipv6_address.split('%')[0]
        ip = ipv4_address or ipv6_address
        client_id = iface.get('client_id')

        if not name:
            LOG.error('Malformed interface record for node %(node)s: %(if)s',
                      {'node': node.uuid, 'if': iface})
            continue

        if not mac:
            LOG.debug('Skipping interface %(if)s for node %(node)s without '
                      'link information',
                      {'node': node.uuid, 'if': iface})
            continue

        if not netutils.is_valid_mac(mac):
            LOG.warning('MAC address of interface %(if)s of node %(node)s '
                        'is not valid, skipping',
                        {'node': node.uuid, 'if': iface})
            continue

        mac = mac.lower()

        LOG.debug('Found interface %(name)s with MAC "%(mac)s", '
                  'IP address "%(ip)s" and client_id "%(client_id)s" '
                  'for node %(node)s',
                  {'name': name, 'mac': mac, 'ip': ip,
                   'client_id': client_id, 'node': node.uuid})
        result[name] = dict(iface, pxe_enabled=(mac == pxe_mac),
                            # IPv6 address without scope.
                            ipv6_address=ipv6_address)

    return result


def validate_interfaces(node, inventory, interfaces):
    """Validate interfaces on correctness and suitability.

    :return: dict interface name -> interface.
    """
    if not interfaces:
        raise exception.InvalidNodeInventory(
            node=node.uuid,
            reason=_('no valid network interfaces'))

    pxe_mac = get_pxe_mac(inventory)
    if not pxe_mac and CONF.inspector.add_ports == 'pxe':
        LOG.warning('No boot interface provided in the inventory for node '
                    '%s, will add all ports with IP addresses', node.uuid)

    result = {}

    for name, iface in interfaces.items():
        ip = iface.get('ipv4_address') or iface.get('ipv6_address')
        pxe = iface.get('pxe_enabled', True)

        if name == 'lo' or (ip and utils.is_loopback(ip)):
            LOG.debug('Skipping local interface %(iface)s for node %(node)s',
                      {'iface': name, 'node': node.uuid})
            continue

        if CONF.inspector.add_ports == 'pxe' and pxe_mac and not pxe:
            LOG.debug('Skipping interface %(iface)s for node %(node)s as it '
                      'was not PXE booting and add_ports is set to "pxe"',
                      {'iface': name, 'node': node.uuid})
            continue

        if CONF.inspector.add_ports == 'active' and not ip:
            LOG.debug('Skipping interface %(iface)s for node %(node)s as it '
                      'did not have an IP address assigned during the ramdisk '
                      'run and add_ports is set to "active"',
                      {'iface': name, 'node': node.uuid})
            continue

        result[name] = iface.copy()

    if not result:
        raise exception.InvalidNodeInventory(
            node=node.uuid,
            reason=_('no network interfaces match the configuration '
                     '(add_ports set to "%s")') % CONF.inspector.add_ports)
    return result


def get_pxe_mac(inventory):
    """Get MAC address of the PXE interface."""
    pxe_mac = inventory.get('boot', {}).get('pxe_interface')
    if pxe_mac and '-' in pxe_mac:
        # pxelinux format: 01-aa-bb-cc-dd-ee-ff
        pxe_mac = pxe_mac.split('-', 1)[1]
        pxe_mac = pxe_mac.replace('-', ':').lower()
    return pxe_mac



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import builtins
import errno
import os
from unittest import mock

import fixtures

from ironic.conf import CONF
from ironic.pxe_filter import dnsmasq
from ironic.tests import base as test_base


class TestExclusiveWriteOrPass(test_base.TestCase):
    def setUp(self):
        super().setUp()
        self.mock_open = self.useFixture(fixtures.MockPatchObject(
            builtins, 'open', new=mock.mock_open())).mock
        self.mock_fd = self.mock_open.return_value
        self.mock_fcntl = self.useFixture(fixtures.MockPatchObject(
            dnsmasq.fcntl, 'flock', autospec=True)).mock
        self.path = '/foo/bar/baz'
        self.buf = 'spam'
        self.fcntl_lock_call = mock.call(
            self.mock_fd, dnsmasq.fcntl.LOCK_EX | dnsmasq.fcntl.LOCK_NB)
        self.fcntl_unlock_call = mock.call(self.mock_fd, dnsmasq.fcntl.LOCK_UN)
        self.mock_log = self.useFixture(fixtures.MockPatchObject(
            dnsmasq.LOG, 'debug')).mock
        self.mock_sleep = self.useFixture(fixtures.MockPatchObject(
            dnsmasq.time, 'sleep')).mock

    def test_write(self):
        wrote = dnsmasq._exclusive_write_or_pass(self.path, self.buf)
        self.assertTrue(wrote)
        self.mock_open.assert_called_once_with(self.path, 'w', 1)
        self.mock_fcntl.assert_has_calls(
            [self.fcntl_lock_call, self.fcntl_unlock_call])
        self.mock_fd.write.assert_called_once_with(self.buf)
        self.mock_log.assert_not_called()

    def test_write_would_block(self):
        # lock/unlock paired calls
        self.mock_fcntl.side_effect = [
            # first try
            BlockingIOError, None,
            # second try
            None, None]
        wrote = dnsmasq._exclusive_write_or_pass(self.path, self.buf)

        self.assertTrue(wrote)
        self.mock_open.assert_called_once_with(self.path, 'w', 1)
        self.mock_fcntl.assert_has_calls(
            [self.fcntl_lock_call, self.fcntl_unlock_call],
            [self.fcntl_lock_call, self.fcntl_unlock_call])
        self.mock_fd.write.assert_called_once_with(self.buf)
        self.mock_log.assert_called_once_with(
            '%s locked; will try again (later)', self.path)
        self.mock_sleep.assert_called_once_with(
            dnsmasq._EXCLUSIVE_WRITE_ATTEMPTS_DELAY)

    @mock.patch.object(dnsmasq, '_EXCLUSIVE_WRITE_ATTEMPTS', 1)
    def test_write_would_block_too_many_times(self):
        self.mock_fcntl.side_effect = [BlockingIOError, None]

        wrote = dnsmasq._exclusive_write_or_pass(self.path, self.buf)
        self.assertFalse(wrote)
        self.mock_open.assert_called_once_with(self.path, 'w', 1)
        self.mock_fcntl.assert_has_calls(
            [self.fcntl_lock_call, self.fcntl_unlock_call])
        self.mock_fd.write.assert_not_called()
        retry_log_call = mock.call('%s locked; will try again (later)',
                                   self.path)
        failed_log_call = mock.call(
            'Failed to write the exclusively-locked path: %(path)s for '
            '%(attempts)s times', {
                'attempts': dnsmasq._EXCLUSIVE_WRITE_ATTEMPTS,
                'path': self.path
            })
        self.mock_log.assert_has_calls([retry_log_call, failed_log_call])
        self.mock_sleep.assert_called_once_with(
            dnsmasq._EXCLUSIVE_WRITE_ATTEMPTS_DELAY)

    def test_write_custom_ioerror(self):

        err = IOError('Oops!')
        err.errno = errno.EBADF
        self.mock_fcntl.side_effect = [err, None]

        self.assertRaisesRegex(
            IOError, 'Oops!', dnsmasq._exclusive_write_or_pass, self.path,
            self.buf)

        self.mock_open.assert_called_once_with(self.path, 'w', 1)
        self.mock_fcntl.assert_has_calls(
            [self.fcntl_lock_call, self.fcntl_unlock_call])
        self.mock_fd.write.assert_not_called()
        self.mock_log.assert_not_called()


class TestHelpers(test_base.TestCase):
    def setUp(self):
        super().setUp()
        self.mac = 'ff:ff:ff:ff:ff:ff'
        self.dhcp_hostsdir = '/far'
        self.path = os.path.join(self.dhcp_hostsdir, self.mac)
        self.unknown_path = os.path.join(self.dhcp_hostsdir,
                                         dnsmasq._UNKNOWN_HOSTS_FILE)
        CONF.set_override('dhcp_hostsdir', self.dhcp_hostsdir,
                          'pxe_filter')
        self.mock__exclusive_write_or_pass = self.useFixture(
            fixtures.MockPatchObject(dnsmasq, '_exclusive_write_or_pass')).mock
        self.mock_stat = self.useFixture(
            fixtures.MockPatchObject(os, 'stat')).mock
        self.mock_listdir = self.useFixture(
            fixtures.MockPatchObject(os, 'listdir')).mock
        self.mock_log = self.useFixture(
            fixtures.MockPatchObject(dnsmasq, 'LOG')).mock

    def test__allowlist_unknown_hosts(self):
        dnsmasq._configure_unknown_hosts(True)

        self.mock__exclusive_write_or_pass.assert_called_once_with(
            self.unknown_path, '%s' % dnsmasq._ALLOW_UNKNOWN_HOSTS)
        self.mock_log.debug.assert_called_once_with(
            'A %s record for all unknown hosts using wildcard mac '
            'created', 'allow')

    def test__denylist_unknown_hosts(self):
        dnsmasq._configure_unknown_hosts(False)

        self.mock__exclusive_write_or_pass.assert_called_once_with(
            self.unknown_path, '%s' % dnsmasq._DENY_UNKNOWN_HOSTS)
        self.mock_log.debug.assert_called_once_with(
            'A %s record for all unknown hosts using wildcard mac '
            'created', 'deny')

    def test__configure_removedlist_allowlist(self):
        self.mock_stat.return_value.st_size = dnsmasq._MAC_DENY_LEN

        dnsmasq._configure_removedlist({self.mac}, True)

        self.mock__exclusive_write_or_pass.assert_called_once_with(
            self.path, '%s\n' % self.mac)

    def test__configure_removedlist_denylist(self):
        self.mock_stat.return_value.st_size = dnsmasq._MAC_ALLOW_LEN

        dnsmasq._configure_removedlist({self.mac}, False)

        self.mock__exclusive_write_or_pass.assert_called_once_with(
            self.path, '%s,ignore\n' % self.mac)

    def test__allowlist_mac(self):
        dnsmasq._add_mac_to_allowlist(self.mac)

        self.mock__exclusive_write_or_pass.assert_called_once_with(
            self.path, '%s\n' % self.mac)

    def test__denylist_mac(self):
        dnsmasq._add_mac_to_denylist(self.mac)

        self.mock__exclusive_write_or_pass.assert_called_once_with(
            self.path, '%s,ignore\n' % self.mac)

    def test__get_denylist(self):
        self.mock_listdir.return_value = [self.mac]
        self.mock_stat.return_value.st_size = len('%s,ignore\n' % self.mac)
        denylist, allowlist = dnsmasq._get_deny_allow_lists()

        self.assertEqual({self.mac}, denylist)
        self.mock_listdir.assert_called_once_with(self.dhcp_hostsdir)
        self.mock_stat.assert_called_with(self.path)

    def test__get_allowlist(self):
        self.mock_listdir.return_value = [self.mac]
        self.mock_stat.return_value.st_size = len('%s\n' % self.mac)
        denylist, allowlist = dnsmasq._get_deny_allow_lists()

        self.assertEqual({self.mac}, allowlist)
        self.mock_listdir.assert_called_once_with(self.dhcp_hostsdir)
        self.mock_stat.assert_called_with(self.path)

    def test__get_no_denylist(self):
        self.mock_listdir.return_value = [self.mac]
        self.mock_stat.return_value.st_size = len('%s\n' % self.mac)
        denylist, allowlist = dnsmasq._get_deny_allow_lists()

        self.assertEqual(set(), denylist)
        self.mock_listdir.assert_called_once_with(self.dhcp_hostsdir)
        self.mock_stat.assert_called_with(self.path)

    def test__get_no_allowlist(self):
        self.mock_listdir.return_value = [self.mac]
        self.mock_stat.return_value.st_size = len('%s,ignore\n' % self.mac)
        denylist, allowlist = dnsmasq._get_deny_allow_lists()

        self.assertEqual(set(), allowlist)
        self.mock_listdir.assert_called_once_with(self.dhcp_hostsdir)
        self.mock_stat.assert_called_with(self.path)


@mock.patch.object(dnsmasq, '_configure_unknown_hosts', autospec=True)
@mock.patch.object(dnsmasq, '_add_mac_to_denylist', autospec=True)
@mock.patch.object(dnsmasq, '_add_mac_to_allowlist', autospec=True)
class TestUpdate(test_base.TestCase):

    def test_no_update(self, mock_allow, mock_deny, mock_configure_unknown):
        dnsmasq.update([], [])
        mock_allow.assert_not_called()
        mock_deny.assert_not_called()
        mock_configure_unknown.assert_not_called()

    def test_only_allow(self, mock_allow, mock_deny, mock_configure_unknown):
        dnsmasq.update(['mac1', 'mac2'], [], allow_unknown=True)
        mock_allow.assert_has_calls([mock.call(f'mac{i}') for i in (1, 2)])
        mock_deny.assert_not_called()
        mock_configure_unknown.assert_called_once_with(True)

    def test_only_deny(self, mock_allow, mock_deny, mock_configure_unknown):
        dnsmasq.update([], ['mac1', 'mac2'])
        mock_allow.assert_not_called()
        mock_deny.assert_has_calls([mock.call(f'mac{i}') for i in (1, 2)])
        mock_configure_unknown.assert_not_called()


@mock.patch.object(dnsmasq, '_configure_removedlist', autospec=True)
@mock.patch.object(dnsmasq, '_configure_unknown_hosts', autospec=True)
@mock.patch.object(dnsmasq, '_add_mac_to_denylist', autospec=True)
@mock.patch.object(dnsmasq, '_add_mac_to_allowlist', autospec=True)
@mock.patch.object(dnsmasq, '_get_deny_allow_lists', autospec=True)
class TestSync(test_base.TestCase):

    def test_no_macs(self, mock_get_lists, mock_allow, mock_deny,
                     mock_configure_unknown, mock_configure_removedlist):
        mock_get_lists.return_value = set(), set()
        dnsmasq.sync([], [], False)
        mock_allow.assert_not_called()
        mock_deny.assert_not_called()
        mock_configure_unknown.assert_called_once_with(False)
        mock_configure_removedlist.assert_called_once_with(set(), False)

    def test_only_new_macs(self, mock_get_lists, mock_allow, mock_deny,
                           mock_configure_unknown, mock_configure_removedlist):
        mock_get_lists.return_value = set(), set()
        dnsmasq.sync(['allow1', 'allow2'], [], True)
        mock_allow.assert_has_calls(
            [mock.call(f'allow{i}') for i in (1, 2)],
            any_order=True)
        mock_deny.assert_not_called()
        mock_configure_unknown.assert_called_once_with(True)
        mock_configure_removedlist.assert_called_once_with(set(), True)

    def test_deny_macs(self, mock_get_lists, mock_allow, mock_deny,
                       mock_configure_unknown, mock_configure_removedlist):
        mock_get_lists.return_value = set(), {'deny1', 'allow1'}
        dnsmasq.sync(['allow1'], ['deny1', 'deny2'], False)
        mock_allow.assert_not_called()
        mock_deny.assert_has_calls(
            [mock.call(f'deny{i}') for i in (1, 2)],
            any_order=True)
        mock_configure_unknown.assert_called_once_with(False)
        mock_configure_removedlist.assert_called_once_with(set(), False)

    def test_removed_nodes(self, mock_get_lists, mock_allow, mock_deny,
                           mock_configure_unknown, mock_configure_removedlist):
        mock_get_lists.return_value = {'mac1'}, {'mac2', 'mac3'}
        dnsmasq.sync(['mac2'], [], True)
        mock_allow.assert_not_called()
        mock_deny.assert_not_called()
        mock_configure_unknown.assert_called_once_with(True)
        mock_configure_removedlist.assert_called_once_with(
            {'mac1', 'mac3'}, True)

    def test_change_state(self, mock_get_lists, mock_allow, mock_deny,
                          mock_configure_unknown, mock_configure_removedlist):
        # MAC1 from denied to allowed, MAC2 from allowed to denied, drop MAC3
        mock_get_lists.return_value = {'mac1'}, {'mac2', 'mac3'}
        dnsmasq.sync(['mac1'], ['mac2'], False)
        mock_allow.assert_called_once_with('mac1')
        mock_deny.assert_called_once_with('mac2')
        mock_configure_unknown.assert_called_once_with(False)
        mock_configure_removedlist.assert_called_once_with({'mac3'}, False)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Systemd Quadlet console container provider.
"""
import json
import os
import re

from oslo_concurrency import processutils
from oslo_log import log as logging

from ironic.common import exception
from ironic.common import utils
from ironic.conf import CONF
from ironic.console.container import base

LOG = logging.getLogger(__name__)

TEMPLATE_PREFIX = 'ironic-console'

# "podman port" output is of the format
# 5900/tcp -> 127.0.0.1:12345
#             ^^^^^^^^^ ^^^^^
PORT_RE = re.compile('^5900/tcp -> (.*):([0-9]+)$')


class SystemdConsoleContainer(base.BaseConsoleContainer):
    """Console container provider which uses Systemd Quadlets."""

    unit_dir = None

    def __init__(self):

        # confirm podman and systemctl are available
        try:
            utils.execute('systemctl', '--version')
        except processutils.ProcessExecutionError as e:
            LOG.exception('systemctl not available, '
                          'this provider cannot be used.')
            raise exception.ConsoleContainerError(provider='systemd', reason=e)
        try:
            utils.execute('podman', '--version')
        except processutils.ProcessExecutionError as e:
            LOG.exception('podman not available, '
                          'it is mandatory to use this provider.')
            raise exception.ConsoleContainerError(provider='systemd', reason=e)

    def _init_unit_dir(self, unit_dir=None):

        if unit_dir:
            self.unit_dir = unit_dir

        elif not self.unit_dir:

            # Write container files to
            # /etc/containers/systemd/users/{uid}/containers/systemd
            # Containers are stateless and can be run rootless as user
            # containers.
            uid = str(os.getuid())
            user_dir = os.path.join('/etc/containers/systemd/users', uid)
            if not os.path.isdir(user_dir):
                reason = (f'Directory {user_dir} must exist and be writable '
                          f'by user {uid}')
                raise exception.ConsoleContainerError(
                    provider='systemd', reason=reason)

            self.unit_dir = os.path.join(
                '/etc/containers/systemd/users', uid, 'containers/systemd')

        if not os.path.exists(self.unit_dir):
            try:
                os.makedirs(self.unit_dir)
            except OSError as e:
                LOG.exception(
                    'Could not initialize console containers')
                raise exception.ConsoleContainerError(
                    provider='systemd', reason=e)

    def _container_path(self, identifier):
        """Build a container path.

        :param identifier: Optional identifier to include in the path
        :returns: A quadlet .container file path
        """
        return os.path.join(
            self.unit_dir, f'{TEMPLATE_PREFIX}-{identifier}.container')

    def _unit_name(self, identifier):
        """Build a unit name.

        :param identifier: Optional identifier to include in the name
        :returns: Unit service name which corresponds to a .container file
        """
        return f'{TEMPLATE_PREFIX}-{identifier}.service'

    def _container_name(self, identifier):
        """Build a container name.

        :param identifier: Optional identifier to include in the name
        :returns: The name of the podman container created by systemd
                  quadlet container
        """
        return f'systemd-{TEMPLATE_PREFIX}-{identifier}'

    def _reload(self):
        """Call systemctl --user daemon-reload

        :raises: ConsoleContainerError
        """
        try:
            utils.execute('systemctl', '--user', 'daemon-reload')
        except processutils.ProcessExecutionError as e:
            LOG.exception('Problem calling systemctl daemon-reload')
            raise exception.ConsoleContainerError(provider='systemd', reason=e)

    def _start(self, unit):
        """Call systemctl --user start.

        :param unit: Name of the unit to start
        :raises: ConsoleContainerError
        """
        try:
            utils.execute('systemctl', '--user', 'start', unit)
        except processutils.ProcessExecutionError as e:
            LOG.exception('Problem calling systemctl start')
            raise exception.ConsoleContainerError(provider='systemd', reason=e)

    def _stop(self, unit):
        """Call systemctl --user stop.

        :param unit: Name of the unit to stop
        :raises: ConsoleContainerError
        """
        try:
            utils.execute('systemctl', '--user', 'stop', unit)
        except processutils.ProcessExecutionError as e:
            LOG.exception('Problem calling systemctl stop')
            raise exception.ConsoleContainerError(provider='systemd', reason=e)

    def _host_port(self, container):
        """Extract running host and port from a container.

        Calls 'podman port' and parses the result.

        :param container: container name
        :returns: Tuple of host IP address and published port
        :raises: ConsoleContainerError
        """
        try:
            out, _ = utils.execute('podman', 'port', container)
            match = PORT_RE.match(out)
            if match:
                return match.group(1), int(match.group(2))
            raise exception.ConsoleContainerError(
                provider='systemd',
                reason=f'Could not detect port in the output: {out}')

        except processutils.ProcessExecutionError as e:
            LOG.exception('Problem calling podman port %s', container)
            raise exception.ConsoleContainerError(provider='systemd', reason=e)

    def _write_container_file(self, identifier, app_name, app_info):
        """Create quadlet container file.

        :param identifier: Unique container identifier
        :param app_name: Sets container environment APP value
        :param app_info: Sets container environment APP_INFO value
        :raises: ConsoleContainerError
        """
        try:
            container_file = self._container_path(identifier)

            # TODO(stevebaker) Support bind-mounting certificate files to
            # handle verified BMC certificates

            params = {
                'description': 'A VNC server which displays a console '
                               f'for node {identifier}',
                'image': CONF.vnc.console_image,
                'port': CONF.vnc.systemd_container_publish_port,
                'app': app_name,
                'app_info': json.dumps(app_info),
                'read_only': CONF.vnc.read_only,
            }

            LOG.debug('Writing %s', container_file)
            with open(container_file, 'wt') as fp:
                fp.write(utils.render_template(
                    CONF.vnc.systemd_container_template, params=params))

        except OSError as e:
            LOG.exception('Could not start console container')
            raise exception.ConsoleContainerError(provider='systemd', reason=e)

    def _delete_container_file(self, identifier):
        """Delete container file.

        :param identifier: Unique container identifier
        :raises: ConsoleContainerError
        """
        container_file = self._container_path(identifier)

        try:
            if os.path.exists(container_file):
                LOG.debug('Removing file %s', container_file)
                os.remove(container_file)
        except OSError as e:
            LOG.exception('Could not stop console containers')
            raise exception.ConsoleContainerError(provider='systemd', reason=e)

    def start_container(self, task, app_name, app_info):
        """Stop a console container for a node.

        Any existing running container for this node will be stopped.

        :param task: A TaskManager instance.
        :raises: ConsoleContainerError
        """
        self._init_unit_dir()
        node = task.node
        uuid = node.uuid

        LOG.debug('Starting console container for node %s', uuid)

        self._write_container_file(
            identifier=uuid, app_name=app_name, app_info=app_info)

        # notify systemd to changed file
        self._reload()

        # start the container
        unit = self._unit_name(uuid)
        try:
            self._start(unit)
        except Exception as e:
            try:
                self._delete_container_file(uuid)
                pass
            except Exception:
                pass
            raise e

        container = self._container_name(uuid)

        return self._host_port(container)

    def _stop_container(self, identifier):
        """Stop a console container for a node.

        Any existing running container for this node will be stopped.

        :param identifier: Unique container identifier
        :raises: ConsoleContainerError
        """
        unit = self._unit_name(identifier)
        try:
            # stop any running container
            self._stop(unit)
        except Exception:
            pass

        self._delete_container_file(identifier)

    def stop_container(self, task):
        """Stop a console container for a node.

        Any existing running container for this node will be stopped.

        :param task: A TaskManager instance.
        :raises: ConsoleContainerError
        """
        self._init_unit_dir()
        node = task.node
        uuid = node.uuid
        LOG.debug('Stopping console container for node %s', uuid)
        self._stop_container(uuid)
        # notify systemd to changed file
        self._reload()

    def stop_all_containers(self):
        """Stops all running console containers

        This is run on conductor startup and graceful shutdown to ensure
        no console containers are running.
        :raises: ConsoleContainerError
        """
        LOG.debug('Stopping all console containers')
        self._init_unit_dir()
        stop_count = 0
        if not os.path.exists(self.unit_dir):
            # No unit state, so assume no containers are running
            return

        for filename in os.listdir(self.unit_dir):
            if not filename.startswith(TEMPLATE_PREFIX):
                # ignore containers this isn't managing
                continue

            stop_count = stop_count + 1
            try:
                # get the identifier from the filename and stop the container
                identifier = filename.split(
                    f'{TEMPLATE_PREFIX}-')[1].split('.container')[0]
                self._stop_container(identifier)
            except Exception:
                pass

        if stop_count > 0:
            try:
                # notify systemd to changed file
                self._reload()
            except Exception:
                pass



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import signal
import time

from oslo_config import cfg
from oslo_log import log
from oslo_utils import timeutils

from ironic.common import console_factory
from ironic.common import rpc
from ironic.common import rpc_service

LOG = log.getLogger(__name__)
CONF = cfg.CONF


class RPCService(rpc_service.BaseRPCService):

    def __init__(self, host, manager_module, manager_class):
        super().__init__(host, manager_module, manager_class)
        self.deregister = True
        self.draining = False

    def _real_start(self):
        super()._real_start()
        rpc.set_global_manager(self.manager)

        # Start in a known state of no console containers running.
        # Any enabled console managed by this conductor will be started
        # after this
        self._stop_console_containers()

    def stop(self):
        initial_time = timeutils.utcnow()
        extend_time = initial_time + datetime.timedelta(
            seconds=CONF.hash_ring_reset_interval)

        try:
            self.manager.del_host(deregister=self.deregister,
                                  clear_node_reservations=False)
        except Exception as e:
            LOG.exception('Service error occurred when cleaning up '
                          'the RPC manager. Error: %s', e)

        if self.manager.get_online_conductor_count() > 1:
            # Delay stopping the server until the hash ring has been
            # reset on the cluster
            stop_time = timeutils.utcnow()
            if stop_time < extend_time:
                stop_wait = max(0, (extend_time - stop_time).seconds)
                LOG.info('Waiting %(stop_wait)s seconds for hash ring reset.',
                         {'stop_wait': stop_wait})
                time.sleep(stop_wait)

        try:
            if self.rpcserver is not None:
                self.rpcserver.stop()
                self.rpcserver.wait()
        except Exception as e:
            LOG.exception('Service error occurred when stopping the '
                          'RPC server. Error: %s', e)

        super().stop(graceful=True)
        LOG.info('Stopped RPC server for service %(service)s on host '
                 '%(host)s.',
                 {'service': self.topic, 'host': self.host})

        # Stop all running console containers
        self._stop_console_containers()

        # Wait for reservation locks held by this conductor.
        # The conductor process will end when one of the following occurs:
        # - All reservations for this conductor are released
        # - shutdown_timeout has elapsed
        # - The process manager (systemd, kubernetes) sends SIGKILL after the
        #   configured timeout period
        while (self.manager.has_reserved()
               and not self._shutdown_timeout_reached(initial_time)):
            LOG.info('Waiting for reserved nodes to clear on host %(host)s',
                     {'host': self.host})
            time.sleep(1)

        # Stop the keepalive heartbeat greenthread sending touch(online=False)
        self.manager.keepalive_halt()

        rpc.set_global_manager(None)

    def _stop_console_containers(self):
        # the default provider is fake, so this can be called even when
        # CONF.vnc.enabled is false
        provider = console_factory.ConsoleContainerFactory().provider
        provider.stop_all_containers()

    def _shutdown_timeout_reached(self, initial_time):
        if self.draining:
            shutdown_timeout = CONF.drain_shutdown_timeout
        else:
            shutdown_timeout = CONF.graceful_shutdown_timeout
        if shutdown_timeout == 0:
            # No timeout, run until no nodes are reserved
            return False
        shutdown_time = initial_time + datetime.timedelta(
            seconds=shutdown_timeout)
        return shutdown_time < timeutils.utcnow()

    def _handle_no_deregister(self, signo, frame):
        LOG.info('Got signal SIGUSR1. Not deregistering on next shutdown '
                 'of service %(service)s on host %(host)s.',
                 {'service': self.topic, 'host': self.host})
        self.deregister = False

    def _handle_drain(self, signo, frame):
        LOG.info('Got signal SIGUSR2. Starting drain shutdown'
                 'of service %(service)s on host %(host)s.',
                 {'service': self.topic, 'host': self.host})
        self.draining = True
        self.stop()

    def handle_signal(self):
        """Add a signal handler for SIGUSR1, SIGUSR2.

        The SIGUSR1 handler ensures that the manager is not deregistered when
        it is shutdown.

        The SIGUSR2 handler starts a drain shutdown.
        """
        signal.signal(signal.SIGUSR1, self._handle_no_deregister)
        signal.signal(signal.SIGUSR2, self._handle_drain)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import uuidutils

from ironic.common import context as ironic_context
from ironic.conductor import base_manager
from ironic.conductor import periodics
from ironic.conductor import task_manager
from ironic.drivers.modules import fake
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils


_FILTERS = {'maintenance': False}


class PeriodicTestService(base_manager.BaseConductorManager):

    def __init__(self, test):
        self.test = test
        self.nodes = []

    @periodics.node_periodic(purpose="herding cats", spacing=42)
    def simple(self, task, context):
        self.test.assertIsInstance(context, ironic_context.RequestContext)
        self.test.assertTrue(task.shared)
        # This may raise
        task.upgrade_lock()
        self.nodes.append(task.node.uuid)

    @periodics.node_periodic(purpose="herding cats", spacing=42,
                             shared_task=False, filters=_FILTERS)
    def exclusive(self, task, context):
        self.test.assertIsInstance(context, ironic_context.RequestContext)
        self.test.assertFalse(task.shared)
        self.nodes.append(task.node.uuid)

    @periodics.node_periodic(purpose="never running", spacing=42,
                             predicate=lambda n: n.cat != 'meow',
                             predicate_extra_fields=['cat'])
    def never_run(self, task, context):
        self.test.fail(f"Was not supposed to run, ran with {task.node}")

    @periodics.node_periodic(purpose="herding cats", spacing=42, limit=3)
    def limit(self, task, context):
        self.test.assertIsInstance(context, ironic_context.RequestContext)
        self.test.assertTrue(task.shared)
        self.nodes.append(task.node.uuid)
        if task.node.uuid == 'stop':
            raise periodics.Stop()


class PeriodicTestInterface(fake.FakePower):

    def __init__(self, test):
        self.test = test
        self.nodes = []

    @periodics.node_periodic(purpose="herding cats", spacing=42)
    def simple(self, task, manager, context):
        self.test.assertIsInstance(manager, PeriodicTestService)
        self.test.assertIsInstance(context, ironic_context.RequestContext)
        self.nodes.append(task.node.uuid)


@mock.patch.object(PeriodicTestService, 'iter_nodes', autospec=True)
class NodePeriodicTestCase(db_base.DbTestCase):

    def setUp(self):
        super().setUp()
        self.service = PeriodicTestService(self)
        self.ctx = ironic_context.get_admin_context()
        self.uuid = uuidutils.generate_uuid()
        self.node = obj_utils.create_test_node(self.context, uuid=self.uuid)

    @mock.patch.object(periodics.LOG, 'info', autospec=True)
    def test_simple(self, mock_log, mock_iter_nodes):
        node2 = obj_utils.create_test_node(self.context,
                                           uuid=uuidutils.generate_uuid(),
                                           reservation='host0')
        mock_iter_nodes.return_value = iter([
            (uuidutils.generate_uuid(), 'driver1', ''),
            (self.uuid, 'driver2', 'group'),
            (node2.uuid, 'driver3', 'group'),
        ])

        self.service.simple(self.ctx)

        mock_iter_nodes.assert_called_once_with(self.service,
                                                filters=None, fields=())
        self.assertEqual([self.uuid], self.service.nodes)
        # 1 node not found, 1 locked
        self.assertEqual(2, mock_log.call_count)

    def test_exclusive(self, mock_iter_nodes):
        mock_iter_nodes.return_value = iter([
            (uuidutils.generate_uuid(), 'driver1', ''),
            (self.uuid, 'driver2', 'group'),
        ])

        self.service.exclusive(self.ctx)

        mock_iter_nodes.assert_called_once_with(self.service,
                                                filters=_FILTERS,
                                                fields=())
        self.assertEqual([self.uuid], self.service.nodes)

    @mock.patch.object(task_manager, 'acquire', autospec=True)
    def test_never_run(self, mock_acquire, mock_iter_nodes):
        mock_iter_nodes.return_value = iter([
            (self.uuid, 'driver2', 'group', 'meow'),
        ])

        self.service.never_run(self.ctx)

        mock_iter_nodes.assert_called_once_with(self.service,
                                                filters=None,
                                                fields=['cat'])
        self.assertEqual([], self.service.nodes)
        mock_acquire.assert_not_called()

    @mock.patch.object(task_manager, 'acquire', autospec=True)
    def test_limit(self, mock_acquire, mock_iter_nodes):
        mock_iter_nodes.return_value = iter([
            (self.uuid, 'driver1', ''),
        ] * 10)
        mock_acquire.return_value.__enter__.return_value.node.uuid = self.uuid

        self.service.limit(self.ctx)

        mock_iter_nodes.assert_called_once_with(self.service,
                                                filters=None, fields=())
        self.assertEqual([self.uuid] * 3, self.service.nodes)

    @mock.patch.object(task_manager, 'acquire', autospec=True)
    def test_stop(self, mock_acquire, mock_iter_nodes):
        mock_iter_nodes.return_value = iter([
            (self.uuid, 'driver1', ''),
        ] * 10)
        mock_acquire.return_value.__enter__.return_value.node.uuid = 'stop'

        self.service.limit(self.ctx)

        mock_iter_nodes.assert_called_once_with(self.service,
                                                filters=None, fields=())
        self.assertEqual(['stop'], self.service.nodes)

    @mock.patch.object(task_manager, 'acquire', autospec=True)
    def test_interface_check(self, mock_acquire, mock_iter_nodes):
        mock_iter_nodes.return_value = iter([
            (uuidutils.generate_uuid(), 'driver1', ''),
            (self.uuid, 'driver2', 'group'),
        ])
        iface = PeriodicTestInterface(self)
        tasks = [
            mock.Mock(spec=task_manager.TaskManager,
                      # This will not match the subclass
                      driver=mock.Mock(power=fake.FakePower())),
            mock.Mock(spec=task_manager.TaskManager,
                      node=self.node,
                      driver=mock.Mock(power=iface)),
        ]
        mock_acquire.side_effect = [
            mock.MagicMock(**{'__enter__.return_value': task})
            for task in tasks
        ]

        iface.simple(self.service, self.context)

        mock_iter_nodes.assert_called_once_with(self.service,
                                                filters=None, fields=())
        self.assertEqual([self.uuid], iface.nodes)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Test class for ironic-api notification utilities."""

from unittest import mock

from oslo_utils import uuidutils

from ironic.api.controllers.v1 import notification_utils as notif_utils
from ironic.objects import fields
from ironic.objects import notification
from ironic.tests import base as tests_base
from ironic.tests.unit.objects import utils as obj_utils


class APINotifyTestCase(tests_base.TestCase):

    def setUp(self):
        super(APINotifyTestCase, self).setUp()
        self.node_notify_mock = mock.Mock()
        self.port_notify_mock = mock.Mock()
        self.chassis_notify_mock = mock.Mock()
        self.portgroup_notify_mock = mock.Mock()
        self.node_notify_mock.__name__ = 'NodeCRUDNotification'
        self.port_notify_mock.__name__ = 'PortCRUDNotification'
        self.chassis_notify_mock.__name__ = 'ChassisCRUDNotification'
        self.portgroup_notify_mock.__name__ = 'PortgroupCRUDNotification'
        _notification_mocks = {
            'chassis': (self.chassis_notify_mock,
                        notif_utils.CRUD_NOTIFY_OBJ['chassis'][1]),
            'node': (self.node_notify_mock,
                     notif_utils.CRUD_NOTIFY_OBJ['node'][1]),
            'port': (self.port_notify_mock,
                     notif_utils.CRUD_NOTIFY_OBJ['port'][1]),
            'portgroup': (self.portgroup_notify_mock,
                          notif_utils.CRUD_NOTIFY_OBJ['portgroup'][1])
        }
        self.addCleanup(self._restore, notif_utils.CRUD_NOTIFY_OBJ.copy())
        notif_utils.CRUD_NOTIFY_OBJ = _notification_mocks

    def _restore(self, value):
        notif_utils.CRUD_NOTIFY_OBJ = value

    def test_common_params(self):
        self.config(host='fake-host')
        node = obj_utils.get_test_node(self.context)
        test_level = fields.NotificationLevel.INFO
        test_status = fields.NotificationStatus.SUCCESS
        notif_utils._emit_api_notification(self.context, node, 'create',
                                           test_level, test_status,
                                           chassis_uuid=None)
        init_kwargs = self.node_notify_mock.call_args[1]
        publisher = init_kwargs['publisher']
        event_type = init_kwargs['event_type']
        level = init_kwargs['level']
        self.assertEqual('fake-host', publisher.host)
        self.assertEqual('ironic-api', publisher.service)
        self.assertEqual('create', event_type.action)
        self.assertEqual(test_status, event_type.status)
        self.assertEqual(test_level, level)

    def test_node_notification(self):
        chassis_uuid = uuidutils.generate_uuid()
        node = obj_utils.get_test_node(self.context,
                                       instance_info={'foo': 'baz'},
                                       driver_info={'param': 104})
        test_level = fields.NotificationLevel.INFO
        test_status = fields.NotificationStatus.SUCCESS
        notif_utils._emit_api_notification(self.context, node, 'create',
                                           test_level, test_status,
                                           chassis_uuid=chassis_uuid)
        init_kwargs = self.node_notify_mock.call_args[1]
        payload = init_kwargs['payload']
        event_type = init_kwargs['event_type']
        self.assertEqual('node', event_type.object)
        self.assertEqual(node.uuid, payload.uuid)
        self.assertEqual({'foo': 'baz'}, payload.instance_info)
        self.assertEqual({'param': 104}, payload.driver_info)
        self.assertEqual(chassis_uuid, payload.chassis_uuid)

    def test_node_notification_mask_secrets(self):
        test_info = {'password': 'secret123', 'some_value': 'fake-value'}
        node = obj_utils.get_test_node(self.context,
                                       driver_info=test_info)
        notification.mask_secrets(node)
        self.assertEqual('******', node.driver_info['password'])
        self.assertEqual('fake-value', node.driver_info['some_value'])

    def test_chassis_notification(self):
        chassis = obj_utils.get_test_chassis(self.context,
                                             extra={'foo': 'boo'},
                                             description='bare01')
        test_level = fields.NotificationLevel.INFO
        test_status = fields.NotificationStatus.SUCCESS
        notif_utils._emit_api_notification(self.context, chassis, 'create',
                                           test_level, test_status)
        init_kwargs = self.chassis_notify_mock.call_args[1]
        payload = init_kwargs['payload']
        event_type = init_kwargs['event_type']
        self.assertEqual('chassis', event_type.object)
        self.assertEqual(chassis.uuid, payload.uuid)
        self.assertEqual({'foo': 'boo'}, payload.extra)
        self.assertEqual('bare01', payload.description)

    def test_port_notification(self):
        node_uuid = uuidutils.generate_uuid()
        portgroup_uuid = uuidutils.generate_uuid()
        port = obj_utils.get_test_port(self.context,
                                       address='11:22:33:77:88:99',
                                       local_link_connection={'a': 25},
                                       extra={'as': 34},
                                       pxe_enabled=False)
        test_level = fields.NotificationLevel.INFO
        test_status = fields.NotificationStatus.SUCCESS
        notif_utils._emit_api_notification(self.context, port, 'create',
                                           test_level, test_status,
                                           node_uuid=node_uuid,
                                           portgroup_uuid=portgroup_uuid)
        init_kwargs = self.port_notify_mock.call_args[1]
        payload = init_kwargs['payload']
        event_type = init_kwargs['event_type']
        self.assertEqual('port', event_type.object)
        self.assertEqual(port.uuid, payload.uuid)
        self.assertEqual(node_uuid, payload.node_uuid)
        self.assertEqual(portgroup_uuid, payload.portgroup_uuid)
        self.assertEqual('11:22:33:77:88:99', payload.address)
        self.assertEqual({'a': 25}, payload.local_link_connection)
        self.assertEqual({'as': 34}, payload.extra)
        self.assertIs(False, payload.pxe_enabled)

    def test_portgroup_notification(self):
        node_uuid = uuidutils.generate_uuid()
        portgroup = obj_utils.get_test_portgroup(self.context,
                                                 address='22:55:88:AA:BB:99',
                                                 name='new01',
                                                 mode='mode2',
                                                 extra={'bs': 11})
        test_level = fields.NotificationLevel.INFO
        test_status = fields.NotificationStatus.SUCCESS
        notif_utils._emit_api_notification(self.context, portgroup, 'create',
                                           test_level, test_status,
                                           node_uuid=node_uuid)
        init_kwargs = self.portgroup_notify_mock.call_args[1]
        payload = init_kwargs['payload']
        event_type = init_kwargs['event_type']
        self.assertEqual('portgroup', event_type.object)
        self.assertEqual(portgroup.uuid, payload.uuid)
        self.assertEqual(node_uuid, payload.node_uuid)
        self.assertEqual(portgroup.address, payload.address)
        self.assertEqual(portgroup.name, payload.name)
        self.assertEqual(portgroup.mode, payload.mode)
        self.assertEqual(portgroup.extra, payload.extra)
        self.assertEqual(portgroup.standalone_ports_supported,
                         payload.standalone_ports_supported)

    @mock.patch('ironic.objects.node.NodeMaintenanceNotification',
                autospec=True)
    def test_node_maintenance_notification(self, maintenance_mock):
        maintenance_mock.__name__ = 'NodeMaintenanceNotification'
        node = obj_utils.get_test_node(self.context,
                                       maintenance=True,
                                       maintenance_reason='test reason')
        test_level = fields.NotificationLevel.INFO
        test_status = fields.NotificationStatus.START
        notif_utils._emit_api_notification(self.context, node,
                                           'maintenance_set',
                                           test_level, test_status)
        init_kwargs = maintenance_mock.call_args[1]
        payload = init_kwargs['payload']
        event_type = init_kwargs['event_type']
        self.assertEqual('node', event_type.object)
        self.assertEqual(node.uuid, payload.uuid)
        self.assertEqual(True, payload.maintenance)
        self.assertEqual('test reason', payload.maintenance_reason)

    @mock.patch.object(notification.NotificationBase, 'emit', autospec=True)
    def test_emit_maintenance_notification(self, emit_mock):
        node = obj_utils.get_test_node(self.context)
        test_level = fields.NotificationLevel.INFO
        test_status = fields.NotificationStatus.START
        notif_utils._emit_api_notification(self.context, node,
                                           'maintenance_set',
                                           test_level, test_status)
        emit_mock.assert_called_once_with(mock.ANY, self.context)
